# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``models.dbsettings`` classes."""

from unittest import TestCase

import transaction

from . import DBUnitTestCase


# =============================================================================
class UModelsDBSettingsDBSettingsNoDb(TestCase):
    """Unit test class for :class:`models.dbsettings.Setting` without DB."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from pyramid.testing import setUp, DummyRequest
        from ..models import DB_METADATA, get_dbengine, get_dbsession_factory
        from ..models import get_tm_dbsession

        self.configurator = setUp(settings={
            'sqlalchemy.url': 'sqlite:///:memory:'})

        DB_METADATA.bind = get_dbengine(self.configurator.get_settings())
        dbsession_factory = get_dbsession_factory(DB_METADATA.bind)

        self.request = DummyRequest(
            dbsession=get_tm_dbsession(dbsession_factory, transaction.manager),
            matched_route=None)

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from pyramid.testing import tearDown
        from ..models import DB_METADATA

        tearDown()
        transaction.manager.abort()
        DB_METADATA.drop_all()

    # -------------------------------------------------------------------------
    def test_exists(self):
        """[u:models.dbsettings.DBSettings.exists] without database"""
        from ..models.dbsettings import DBSettings

        # pylint: disable = no-member
        self.assertFalse(DBSettings.exists(self.request.dbsession))

    # -------------------------------------------------------------------------
    def test_db2dict(self):
        """[u:models.dbsettings.DBSettings.db2dict] without database"""
        from ..models.dbsettings import DBSettings

        settings = {'site.uid': 'testchrysalio'}
        # pylint: disable = no-member
        settings_dict = DBSettings.db2dict(
            settings, self.request.dbsession, 'admin@chrysal.io')
        # pylint: enable = no-member
        self.assertIsInstance(settings_dict, dict)
        self.assertIn('title', settings_dict)
        self.assertEqual(settings_dict['title'], 'testchrysalio')


# =============================================================================
class UModelsDBSettingsDBSettings(DBUnitTestCase):
    """Unit test class for :class:`models.dbsettings.Setting`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from pkg_resources import get_distribution

        super(UModelsDBSettingsDBSettings, self).setUp()

        self.request.registry.settings['site.uid'] = 'testchrysalio'
        self.request.registry['version'] = 'chrysalio {0}'.format(
            get_distribution('chrysalio').version)
        self.request.registry['settings']['populate'] = 'now'

    # -------------------------------------------------------------------------
    def test_exists(self):
        """[u:models.dbsettings.Setting.exists]"""
        from ..models.dbsettings import DBSettings

        dbsession = self.request.dbsession

        self.assertFalse(DBSettings.exists(dbsession))

        dbsession.add(DBSettings(key='populate', value='now'))
        self.assertTrue(DBSettings.exists(dbsession))

    # -------------------------------------------------------------------------
    def test_xml2db(self):
        """[u:models.dbsettings.DBSettings.xml2db]"""
        from lxml import etree
        from ..models.dbsettings import DBSettings

        settings_elt = etree.XML(
            '<settings>'
            '  <title>Chrysalio – Test 1</title>'
            '  <email>admin@chrysal.io</email>'
            '  <password-min-length>8</password-min-length>'
            '  <remember-me>5184000</remember-me>'
            '  <page-size>80</page-size>'
            '  <download-max-size>10485760</download-max-size>'
            '  <theme>Default</theme>'
            '</settings>')
        dbsession = self.request.dbsession

        DBSettings.xml2db(dbsession, None)
        self.assertIsNone(dbsession.query(DBSettings).first())

        DBSettings.xml2db(dbsession, settings_elt)
        settings = dbsession.query(DBSettings).all()
        self.assertEqual(len(settings), 8)
        keys = [k.key for k in settings]
        self.assertIn('title', keys)
        self.assertIn('email', keys)
        self.assertIn('password-min-length', keys)
        self.assertIn('remember-me', keys)
        self.assertIn('page-size', keys)
        self.assertIn('download-max-size', keys)
        self.assertIn('theme', keys)
        self.assertIn('populate', keys)

        settings_elt = etree.XML(
            '<settings>'
            '  <title>Chrysalio – Test 2</title>'
            '  <password-min-length>8</password-min-length>'
            '  <page-size>80</page-size>'
            '</settings>')
        DBSettings.xml2db(dbsession, settings_elt)
        settings = dbsession.query(DBSettings).all()
        self.assertEqual(len(settings), 4)

        settings_elt = etree.XML(
            '<settings>'
            '  <title>Chrysalio – Test 3</title>'
            '  <password-min-length>8</password-min-length>'
            '  <page-size>80</page-size>'
            '  <foo bar="baz"/>'
            '</settings>')
        DBSettings.xml2db(dbsession, settings_elt)
        settings = dbsession.query(DBSettings).all()
        self.assertEqual(len(settings), 5)

    # -------------------------------------------------------------------------
    def test_db2xml(self):
        """[u:models.dbsettings.DBSettings.db2xml]"""
        from ..models.dbsettings import DBSettings

        dbsession = self.request.dbsession
        dbsession.add(DBSettings(key='title', value='Chrysalio – Test'))
        dbsession.add(DBSettings(key='password-min-length', value='8'))

        settings_elt = DBSettings.db2xml(dbsession)
        self.assertEqual(settings_elt.findtext('title'), 'Chrysalio – Test')
        self.assertEqual(settings_elt.findtext('password-min-length'), '8')

    # -------------------------------------------------------------------------
    def test_db2dict(self):
        """[u:models.dbsettings.DBSettings.db2dict]"""
        from ..models.dbsettings import DBSettings

        settings = {
            'site.uid': 'testchrysalio', 'site.title': 'Chrysalio – Test 1',
            'pyramid.default_locale_name': 'en'}
        dbsession = self.request.dbsession

        settings_dict = DBSettings.db2dict(
            settings, dbsession, 'admin@chrysal.io')
        self.assertIsInstance(settings_dict, dict)
        self.assertIn('title', settings_dict)
        self.assertIn('password-min-length', settings_dict)
        self.assertIn('email', settings_dict)
        self.assertIn('populate', settings_dict)
        self.assertEqual(settings_dict['title'], 'Chrysalio – Test 1')
        self.assertEqual(settings_dict['password-min-length'], 8)
        self.assertEqual(settings_dict['email'], 'admin@chrysal.io')

        dbsession.add(DBSettings(key='title', value='Chrysalio – Test 2'))
        dbsession.add(DBSettings(key='password-min-length', value='10'))
        settings_dict = DBSettings.db2dict(
            settings, dbsession, 'admin@chrysal.io')
        self.assertEqual(settings_dict['title'], 'Chrysalio – Test 2')
        self.assertEqual(settings_dict['password-min-length'], 10)

    # -------------------------------------------------------------------------
    def test_tab4view(self):
        """[u:models.dbsettings.DBSettings.tab4view]"""
        from ..lib.form import Form
        from ..models.dbsettings import SETTINGS_DEFAULTS, DBSettings

        html = DBSettings.tab4view(self.request, 0, Form(self.request))
        self.assertIn(
            '<div>{0}</div>'.format(SETTINGS_DEFAULTS['page-size']), html)
        self.assertIn(
            '<div>{0}</div>'.format(SETTINGS_DEFAULTS['password-min-length']),
            html)

        html = DBSettings.tab4view(self.request, 1, Form(self.request))
        self.assertNotIn('You cannot modify', html)
        self.assertIn('<div>testchrysalio</div>', html)
        self.assertIn('<div>en</div>', html)
        self.assertIn('<div>sqlite</div>', html)
        self.assertIn('<div>:memory:</div>', html)

        self.assertEqual(
            DBSettings.tab4view(self.request, 2, Form(self.request)), '')

    # -------------------------------------------------------------------------
    def test_settings_schema(self):
        """[u:models.dbsettings.DBSettings.settings_schema]"""
        from colander import SchemaNode
        from ..models.dbsettings import DBSettings

        self.request.registry['themes'] = {'': {}, 'Alternative': {}}
        self.request.registry['settings']['title'] = 'testchrysalio'

        schema, defaults = DBSettings.settings_schema(self.request)
        self.assertIsInstance(schema, SchemaNode)
        serialized = schema.serialize()
        self.assertIn('title', serialized)
        self.assertIn('email', serialized)
        self.assertIn('language', serialized)
        self.assertIn('password-min-length', serialized)
        self.assertIn('page-size', serialized)
        self.assertIn('download-max-size', serialized)
        self.assertIn('theme', serialized)
        self.assertIn('title', defaults)

    # -------------------------------------------------------------------------
    def test_tab4edit(self):
        """[u:models.dbsettings.DBSettings.tab4edit]"""
        from ..lib.form import Form
        from ..models.dbsettings import DBSettings

        self.request.registry['themes'] = {
            '': {'name': {}}, 'Alternative': {'name': {'en': 'Alternative'}}}

        html = DBSettings.tab4edit(self.request, 0, Form(self.request))
        self.assertIn('<h3>Site</h3>', html)
        self.assertIn('Title:', html)
        self.assertIn('Email:', html)
        self.assertIn('<h3>Default values</h3>', html)
        self.assertIn('Language:', html)
        self.assertIn('Password length:', html)
        self.assertIn('Lines per page:', html)
        self.assertNotIn('Themes:', html)

        html = DBSettings.tab4edit(self.request, 1, Form(self.request))
        self.assertIn('You cannot modify', html)

        self.assertEqual(
            DBSettings.tab4edit(self.request, 2, Form(self.request)), '')



"""Console scripts."""

from sys import exit as sys_exit
from logging import getLogger
from os.path import exists, abspath, dirname
from argparse import ArgumentParser
from collections import OrderedDict
from configparser import ConfigParser
from importlib import import_module

from sqlalchemy.exc import OperationalError
from plaster.exceptions import InvalidURI

from pyramid.paster import get_appsettings

from ..security import PRINCIPALS
from ..lib.i18n import _, translate
from ..lib.config import config_get, settings_get_list
from ..lib.log import setup_logging
from ..lib.utils import tounicode
from ..modules import Module
from ..models import DB_METADATA
from ..models import get_dbengine, get_dbsession_factory


LOG = getLogger(__name__)


# =============================================================================
class ScriptRegistry(dict):
    """Class to simulate Application Registry."""

    def __init__(self, settings):
        """Constructor method."""
        super(ScriptRegistry, self).__init__()
        self.settings = settings


# =============================================================================
class Script(object):
    """Base class for a script which reads the INI file and interacts with the
    database.

    :type  args: argparse.Namespace
    :param args:
        Command line arguments.
    :param bool create_all:
        If ``True``, it creates all the tables, otherwise, it checks if the
        table ``users`` exists with at least one user.
    :param list includes: (optional)
        List of hard coding `includes`.
    :type  dbsession_factory: sqlalchemy.orm.session.sessionmaker
    :param dbsession_factory: (optional)
        Function to create session.
    """

    # -------------------------------------------------------------------------
    def __init__(
            self, args, create_all, includes=None, dbsession_factory=None):
        """Constructor method."""
        self._args = args
        self.registry = None

        try:
            settings = get_appsettings(args.conf_uri, options=args.options)
        except (LookupError, IOError, InvalidURI) as error:
            LOG.critical(error)
            return

        config_file = abspath(args.conf_uri.partition('#')[0])
        self._config = ConfigParser({'here': dirname(config_file)})
        self._config.read(tounicode(config_file), encoding='utf8')

        self.registry = ScriptRegistry(settings)
        self.registry.settings['__file__'] = config_file
        self.registry['principals'] = list(PRINCIPALS)
        self.registry['modules'] = OrderedDict()
        self._load_includes(includes)

        if not self._initialize_database(create_all, dbsession_factory):
            self.registry = None
            return

    # -------------------------------------------------------------------------
    @classmethod
    def argument_parser(cls, description):
        """Create an argument parser object to parse command line arguments.

        :param str description:
            Description of the script.
        :rtype: argparse.ArgumentParser
        """
        parser = ArgumentParser(description=description)
        parser.add_argument(
            'conf_uri', help='URI of configuration (e.g. production.ini#foo)')
        parser.add_argument(
            '--options', dest='options', help='optional configuration options')
        parser.add_argument('--lang', dest='lang', help='user language')
        parser.add_argument(
            '--log-level', dest='log_level', help='log level', default='INFO',
            choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'))
        parser.add_argument('--log-file', dest='log_file', help='log file')
        return parser

    # -------------------------------------------------------------------------
    @classmethod
    def arguments(cls, parser, args=None, filemode='w'):
        """Retrieve arguments from parser, check them and setup logging.

        :type parser: argparse.ArgumentParser
        :param parser:
             Configuration parser object.
        :param list args: (optional)
             Command line arguments (for testing).
        :param str filemode:
             File mode for logging (``'w'`` or ``'a'``).
        :rtype: argparse.Namespace
        """
        args = parser.parse_args(args)
        if not exists(args.conf_uri.partition('#')[0]):
            parser.print_usage()
            return None
        setup_logging(args.log_level, args.log_file, filemode=filemode)
        return args

    # -------------------------------------------------------------------------
    def _load_includes(self, includes=None):
        """Load needed modules with an `includeme()` function and fill
        ``self.registry`` dictionary.

        :param list includes: (optional)
            List of hard coding `includes`.
        """
        # Read include list
        if includes is None:
            includes = []
        includes += [k for k in settings_get_list(
            self.registry.settings, 'chrysalio.includes') if k]

        # Load modules
        for include in includes:
            try:
                module = import_module(include)
            except ImportError as error:
                sys_exit(self._translate(
                    _('*** Unable to load module "${m}" [${e}]', {
                        'm': include, 'e': error})))
            try:
                module.includeme(self.registry)
            except AttributeError as error:
                sys_exit(self._translate(_(
                    '*** Module "${m}" error: ${e}',
                    {'m': include, 'e': error})))

        # Check conflicts
        implementations, error = Module.check_conflicts(
            includes, self.registry['modules'])
        if error is not None:
            sys_exit(self._translate(error))

        # Check dependencies
        for module_id in self.registry['modules']:
            error = self.registry['modules'][module_id].check_dependencies(
                implementations)
            if error is not None:
                sys_exit(self._translate(error))

    # -------------------------------------------------------------------------
    def _initialize_database(self, create_all, dbsession_factory=None):
        """Database initialization.

        :param bool create_all:
            If ``True``, it creates all the tables, otherwise, it checks if the
            table ``users`` exists with at least one user.
        :type  dbsession_factory: sqlalchemy.orm.session.sessionmaker
        :param dbsession_factory: (optional)
            Function to create session.
        :rtype: bool
        """
        # Get database session factory
        self.registry['dbsession_factory'] = dbsession_factory
        if self.registry['dbsession_factory'] is None:
            try:
                dbengine = get_dbengine(self.registry.settings)
            except KeyError:
                LOG.critical(self._translate(_('Database is not defined.')))
                return False
            DB_METADATA.bind = dbengine
            self.registry['dbsession_factory'] = get_dbsession_factory(
                dbengine)

        # Possibly, drop any existing tables
        if hasattr(self._args, 'drop_tables') and (
                self._args.drop_tables or
                config_get(self._config, 'Populate', 'drop_tables') == 'true'):
            LOG.info(self._translate(_('====== Dropping existing tables')))
            tables = set(DB_METADATA.tables)
            try:
                DB_METADATA.reflect()
                DB_METADATA.drop_all()
            except UnicodeEncodeError:  # pragma: nocover
                LOG.error(self._translate(_('Unknown database!')))
                return False
            except OperationalError as error:  # pragma: nocover
                LOG.error(error.args[0])
                return False
            for table in set(DB_METADATA.tables) - tables:  # pragma: nocover
                DB_METADATA.remove(DB_METADATA.tables[table])

        if create_all:
            # Create the tables if they don't already exist
            try:
                DB_METADATA.create_all()
            except OperationalError as error:  # pragma: nocover
                LOG.error(error.args[0])
                return False

        return True

    # -------------------------------------------------------------------------
    def _translate(self, text):
        """Return ``text`` translated.

        :param str text:
            Text to translate.
        :rtype: str
        """
        return translate(text, self._args.lang)



"""Various functions to manage logs."""

from os import makedirs
from os.path import expanduser, dirname, exists
from logging import getLogger, basicConfig
from logging.config import fileConfig
from json import loads

from .config import config_get_list

LOG_ACTIVITY = getLogger('activity')


# =============================================================================
def setup_logging(
        log_level='INFO', log_file=None, log_format=None, filemode='w'):
    """Initialize logging system.

    :param str log_level: ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        Log level. By default `INFO`.
    :param str log_file: (optional)
        Path to log file.
    :param str log_format: (optional)
        Format for log entry. By default,
        `%(asctime)s %(levelname)-8s %(message)s`
    :param str filemode: (default='w')
        File mode: ``'w'`` or ``'a'``.
    """
    if log_format is None:
        log_format = '%(asctime)s %(levelname)-8s %(message)s'
    if log_file:
        basicConfig(
            filename=expanduser(log_file), filemode=filemode, level=log_level,
            format=log_format)
    else:
        basicConfig(level=log_level, format=log_format)


# =============================================================================
def log_activity_setup(config, config_file=None):
    """Create log directory.

    :type  config: configparser.ConfigParser
    :param config:
        Configuration of the site.
    :param str config_file:
        Absolute path to the configuration file.
    :rtype: logging.Logger
    :return:
        Return ``LOG_ACTIVITY`` if activity log is activated.
    """
    if 'activity' not in config_get_list(config, 'loggers', 'keys') or \
       not config.has_option('handler_activity', 'args'):
        return None

    # Setup logging
    if config_file is not None:  # pragma: nocover
        fileConfig(config_file, defaults={'here': dirname(config_file)})

    # Make log directory
    log_args = loads('[{0}]'.format(
        config.get('handler_activity', 'args')[1:-1]
        .replace("'", '"').replace('True', 'true') .replace('False', 'false')))
    path = dirname(log_args[0])
    if not exists(path):
        makedirs(path)

    return LOG_ACTIVITY


# =============================================================================
def log_info(request, action, *args):
    """Write an information message in the log.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str action:
        Action in action.
    :param list args:
        Non-keyworded arguments.
    """
    if request.registry.get('log_activity') is not None:
        request.registry['log_activity'].info(
            '[%s] %s %s',
            (hasattr(request, 'session') and 'user' in request.session and
             request.session['user']['login']) or
            ('context' in request.registry and
             request.registry['context'].get('login')) or '',
            action, ' '.join(args))


# =============================================================================
def log_error(request, error):
    """Write an error message in the log.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    if request.registry.get('log_activity') is not None:
        request.registry['log_activity'].error(
            '[%s] %s',
            (hasattr(request, 'session') and 'user' in request.session and
             request.session['user']['login']) or
            ('context' in request.registry and
             request.registry['context'].get('login')) or '', error)


# =============================================================================
def log_warning(request, warning):
    """Write a warning message in the log.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    if request.registry.get('log_activity') is not None:
        request.registry['log_activity'].warning(
            '[%s] %s',
            (hasattr(request, 'session') and 'user' in request.session and
             request.session['user']['login']) or
            ('context' in request.registry and
             request.registry['context'].get('login')) or '', warning)


# =============================================================================
def log_debug(request, debug):
    """Write a debug message in the log.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    if request.registry.get('log_activity') is not None:
        request.registry['log_activity'].debug(
            '[%s] %s',
            (hasattr(request, 'session') and 'user' in request.session and
             request.session['user']['login']) or
            ('context' in request.registry and
             request.registry['context'].get('login')) or '', debug)



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``models.populate`` functions."""

from os.path import exists, join, dirname
from collections import OrderedDict
from cgi import FieldStorage
from datetime import datetime
from json import dumps

from lxml import etree

from pyramid.response import Response, FileResponse

from . import DBUnitTestCase
from ..lib.xml import relaxng4validation
# pylint: disable = unused-import
from ..includes.modules.models import DBModule  # noqa
from ..modules.cioskeleton.models.dbbone import DBBone  # noqa
# pylint: enable = unused-import


# =============================================================================
class UModelsPopulateXml2Db(DBUnitTestCase):
    """Unit test class for :func:`models.populate.xml2db`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.populate.xml2db]"""
        from . import TEST_INI
        from ..relaxng import RELAXNG
        from ..lib.xml import load_xml
        from ..modules import Module
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUser
        from ..models.populate import xml2db

        dbsession = self.request.dbsession
        data = (
            '<chrysalio version="1.0">'
            '<profiles>'
            '  <profile id="user_creator">'
            '    <label xml:lang="en">User manager</label>'
            '    <principals>'
            '      <principal>user.creator</principal>'
            '    </principals>'
            '  </profile>'
            '</profiles>'
            '<users>'
            '  <user>'
            '    <login>test1</login>'
            '    <password>test1pwd</password>'
            '    <firstname>Édith</firstname>'
            '    <lastname>AVULEUR</lastname>'
            '    <email>test1@chrysal.io</email>'
            '    <profiles>'
            '      <profile>user_creator</profile>'
            '    </profiles>'
            '  </user>'
            '</users>'
            '<groups>'
            '  <group id="team">'
            '    <label xml:lang="en">Team 1</label>'
            '    <users>'
            '      <user>test1</user>'
            '    </users>'
            '  </group>'
            '</groups>'
            '</chrysalio>')
        tree = load_xml(
            'chrysalio.xml', relaxng4validation(RELAXNG),
            data=bytes(data.encode('utf8')))

        errors = xml2db(
            dbsession, tree, modules=OrderedDict(
                [('chrysalio.modules.foo', Module(TEST_INI))]))
        self.assertEqual(len(errors), 0)
        dbprofile = dbsession.query(DBProfile).filter_by(
            profile_id='user_creator').first()
        self.assertIsInstance(dbprofile, DBProfile)
        user = dbsession.query(DBUser).filter_by(login='test1').first()
        self.assertIsInstance(user, DBUser)
        self.assertEqual(len(user.profiles), 1)
        self.assertEqual(len(user.groups), 1)


# =============================================================================
class UModelsPopulateElement2Db(DBUnitTestCase):
    """Unit test class for :func:`models.populate.element2db`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.populate.element2db]"""
        from ..lib.i18n import translate
        from ..models.dbuser import DBUser
        from ..models.populate import element2db

        class DummyDBBone(object):
            """Dummy class for bones."""
            @classmethod
            def xml2db(cls, dbsession, product_elt, error_if_exists=True,
                       kwargs=None):
                """Dummy loading of a product from a XML element."""

        dbsession = self.request.dbsession
        tree = etree.ElementTree(etree.XML(
            '<chrysalio version="1.0">'
            '  <user>'
            '    <login>test1</login>'
            '    <password>test1pwd</password>'
            '    <firstname>Édith</firstname>'
            '    <lastname>AVULEUR</lastname>'
            '    <email>test1@chrysal.io</email>'
            '  </user>'
            '</chrysalio>'))

        # only='profile'
        element = {'tag': 'user', 'class': DBUser}
        self.assertEqual(
            len(element2db(dbsession, tree, 'profile', True, element)), 0)

        # Normal loading
        errors = element2db(dbsession, tree, None, True, element)
        self.assertEqual(len(errors), 1)
        self.assertFalse(errors[0])

        # Module loading
        tree = etree.ElementTree(etree.XML(
            '<chrysalio version="1.0">'
            '  <modules>'
            '    <module id="chrysalio.modules.cioskeleton" active="true">'
            '      <cioskeleton xmlns="http://ns.chrysal.io/cioskeleton"'
            '          version="1.0">'
            '        <bones>'
            '          <bone id="hand">'
            '            <label>Hand</label>'
            '          </bone>'
            '        </bones>'
            '      </cioskeleton>'
            '      <cioskeleton xmlns="http://ns.chrysal.io/cioskeleton"'
            '          version="1.0">'
            '        <bones>'
            '          <bone id="foot">'
            '            <label>Foot</label>'
            '            <foo/>'
            '          </bone>'
            '        </bones>'
            '      </cioskeleton>'
            '    </module>'
            '  </modules>'
            '</chrysalio>'))
        element['relaxng'] = {
            'namespace': 'http://ns.chrysal.io/cioskeleton',
            'root': 'cioskeleton', 'version': '1.0',
            'file': join(
                dirname(__file__), '..', 'modules', 'cioskeleton', 'RelaxNG',
                'cioskeleton.rng')}
        element['tag'] = 'bone'
        element['class'] = DummyDBBone
        errors = element2db(dbsession, tree, None, True, element)
        self.assertEqual(len(errors), 1)
        self.assertIn('foo', translate(errors[0]))


# =============================================================================
class UModelsPopulateModuleXml2Db(DBUnitTestCase):
    """Unit test class for :func:`models.populate.module_xml2db`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.populate.module_xml2db]"""
        from . import TEST_INI
        from ..modules.cioskeleton import ModuleCioSkeleton
        from ..models.populate import module_xml2db

        def _xml2db(dbsession, tree, only, error_if_exists):
            """Dummy function to generate error."""
            # pylint: disable = unused-argument
            return ['Error']

        dbsession = self.request.dbsession

        tree = etree.ElementTree(etree.XML(
            '<chrysalio version="1.0">'
            '  <user>'
            '    <login>test1</login>'
            '    <password>test1pwd</password>'
            '    <firstname>Édith</firstname>'
            '    <lastname>AVULEUR</lastname>'
            '    <email>test1@chrysal.io</email>'
            '  </user>'
            '</chrysalio>'))

        # Without module
        self.assertEqual(
            len(module_xml2db(dbsession, tree, None, True, None)), 0)

        # Module loading
        skeleton = ModuleCioSkeleton(TEST_INI)
        tree = etree.ElementTree(etree.XML(
            '<chrysalio version="1.0">'
            '  <module id="chrysalio.modules.cioskeleton">'
            '    <cioskeleton xmlns="http://ns.chrysal.io/cioskeleton"'
            '        version="1.0">'
            '      <bones>'
            '        <bone id="foot"><label>Foot</label></bone>'
            '      </bones>'
            '    </cioskeleton>'
            '  </module>'
            '</chrysalio>'))
        errors = module_xml2db(
            dbsession, tree, None, True,
            OrderedDict([('chrysalio.modules.cioskeleton', skeleton)]))
        self.assertFalse(errors)

        # Error during module loading
        skeleton.xml2db = (_xml2db,)
        errors = module_xml2db(
            dbsession, tree, None, True,
            OrderedDict([('chrysalio.modules.cioskeleton', skeleton)]))
        self.assertTrue(errors)
        self.assertIsNotNone(errors[0])


# =============================================================================
class UModelsPopulateDb2Xml(DBUnitTestCase):
    """Unit test class for func:`models.populate.db2xml`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.populate.db2xml]"""
        from . import TEST_INI
        from ..modules import Module
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUser
        from ..models.dbgroup import DBGroup, DBGroupUser
        from ..models.populate import db2xml

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_creator',
            i18n_label=dumps({'en': 'User account manager'})))
        self.add_user({
            'login': 'admin', 'status': 'administrator',
            'last_name': 'Administrator', 'password': 'adminpwd',
            'email': 'admin@chrysal.io'})
        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        dbuser = dbsession.query(DBUser).filter_by(login='test1').first()
        dbgroup = DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'}))
        dbgroup.users.append(DBGroupUser(user_id=dbuser.user_id))
        dbsession.add(dbgroup)
        elements = db2xml(
            dbsession,
            modules=OrderedDict([('chrysalio.modules.foo', Module(TEST_INI))]))

        self.assertEqual(len(elements), 3)


# =============================================================================
class UModelsPopulateModuleDb2Xml(DBUnitTestCase):
    """Unit test class for func:`models.populate.module_db2xml`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.populate.module_db2xml]"""
        from . import TEST_INI
        from ..modules.cioskeleton import ModuleCioSkeleton
        from ..models.populate import module_db2xml

        dbsession = self.request.dbsession

        self.assertEqual(len(module_db2xml(dbsession, None)), 0)

        skeleton = ModuleCioSkeleton(TEST_INI)
        self.assertEqual(
            len(module_db2xml(
                dbsession, OrderedDict(
                    [('chrysalio.modules.cioskeleton', skeleton)]))), 1)


# =============================================================================
class UModelsPopulateWeb2Db(DBUnitTestCase):
    """Unit test class for func:`models.populate.web2db`."""

    # -------------------------------------------------------------------------
    def test_xml(self):
        """[u:models.populate.web2db] XML"""
        from . import TEST_DIR, TEST_INI, TEST1_USR_XML, TEST3_USR_XML
        from . import DummyPOST
        from ..relaxng import RELAXNG
        from ..models.populate import web2db, xml2db

        self.request.registry.settings['attachments'] = TEST_DIR
        self.request.registry['relaxng'] = RELAXNG
        self.request.POST = DummyPOST()
        web2db(self.request, xml2db)
        self.assertFalse(self.request.session.pop_flash('alert'))

        # Without file
        self.request.POST.multikeys = {'file': ('foo',)}
        web2db(self.request, xml2db)
        self.assertFalse(self.request.session.pop_flash('alert'))

        # Non well formed XML
        with open(TEST1_USR_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_USR_XML
            input_file.file = hdl
            self.request.POST.multikeys = {'file': (input_file,)}
            web2db(self.request, xml2db)
        self.assertTrue(self.request.session.pop_flash('alert'))

        # Not a XML or ZIP file
        with open(TEST_INI, 'r') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST_INI
            input_file.file = hdl
            self.request.POST.multikeys = {'file': (input_file,)}
            web2db(self.request, xml2db)
        self.assertTrue(self.request.session.pop_flash('alert'))

        # Valid XML
        with open(TEST3_USR_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST3_USR_XML
            input_file.file = hdl
            self.request.POST.multikeys = {'file': (input_file,)}
            web2db(self.request, xml2db)
        self.assertFalse(self.request.session.pop_flash('alert'))

        # Valid but existing XML
        with open(TEST3_USR_XML, 'r') as hdl:
            input_file = FieldStorage()
            self.request.POST.multikeys = {'file': (input_file,)}
            input_file.filename = TEST3_USR_XML
            input_file.file = hdl
            web2db(self.request, xml2db)
        self.assertTrue(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_zip(self):
        """[u:models.populate.web2db] ZIP"""
        from . import TEST_DIR, TEST1_USR_ZIP, TEST2_USR_ZIP, TEST3_USR_ZIP
        from . import DummyPOST
        from ..relaxng import RELAXNG
        from ..models.dbuser import DBUser
        from ..models.populate import web2db, xml2db

        self.request.registry.settings['attachments'] = TEST_DIR
        self.request.registry['relaxng'] = RELAXNG
        self.request.POST = DummyPOST()
        web2db(self.request, xml2db)
        self.assertFalse(self.request.session.pop_flash('alert'))

        with open(TEST2_USR_ZIP, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST2_USR_ZIP
            self.request.POST.multikeys = {'file': (input_file,)}
            web2db(self.request, xml2db)
        self.assertTrue(self.request.session.pop_flash('alert'))

        with open(TEST3_USR_ZIP, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST3_USR_ZIP
            self.request.POST.multikeys = {'file': (input_file,)}
            web2db(self.request, xml2db)
        self.assertTrue(self.request.session.pop_flash('alert'))

        with open(TEST1_USR_ZIP, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST1_USR_ZIP
            self.request.POST.multikeys = {'file': (input_file,)}
            web2db(self.request, xml2db)
        self.assertFalse(self.request.session.pop_flash('alert'))
        self.assertTrue(exists(join(
            TEST_DIR, DBUser.attachments_dir, 'User3', 'user3.svg')))


# =============================================================================
class UModelsPopulateDb2Web(DBUnitTestCase):
    """Unit test class for func:`models.populate.db2web`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.populate.db2web]"""
        from . import ATTACHMENTS_DIR
        from ..relaxng import RELAXNG
        from ..models.populate import db2web
        from ..models.dbuser import DBUser

        user = DBUser(
            status='active', login='test1', first_name='Édith',
            last_name='AVULEUR', honorific='Mrs', email='test1@chrysal.io',
            account_creation=datetime.now())
        user.set_password('sesame')
        self.request.registry.settings = {'site.uid': 'testchrysalio'}
        self.request.registry['relaxng'] = RELAXNG

        response = db2web(self.request, [user], 'users.usr.xml')
        self.assertIsInstance(response, Response)
        self.assertIn('<login>test1</login>', response.text)
        self.assertIn('<firstname>Édith</firstname>', response.text)
        self.assertIn('<lastname>AVULEUR</lastname>', response.text)
        self.assertIn('<honorific>Mrs</honorific>', response.text)
        self.assertIn('<email>test1@chrysal.io</email>', response.text)

        self.request.registry.settings['attachments'] = ATTACHMENTS_DIR
        response = db2web(self.request, [user], 'users.usr.xml')
        self.assertIsInstance(response, Response)

        user = DBUser(
            status='active', login='test1', first_name='Édith',
            last_name='AVULEUR', honorific='Mrs', email='test1@chrysal.io',
            attachments_key='Test1', picture='test1.svg',
            account_creation=datetime.now())
        user.set_password('sesame')

        response = db2web(self.request, [user], 'users.usr.xml')
        self.assertIsInstance(response, FileResponse)



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``views.user`` methods."""

from ..models.dbgroup import DBGroup
from . import DBUnitTestCase, FunctionalTestCase


# =============================================================================
class UViewsUserUserView(DBUnitTestCase):
    """Unit test class for testing :class:`views.user.UserView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""

        super(UViewsUserUserView, self).setUp()

        self.add_user({
            'login': 'admin', 'status': 'administrator', 'last_name': 'ADMIN',
            'password': 'adminpwd', 'email': 'admin@chrysal.io'})
        self.add_user({
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        self.add_user({
            'login': 'test2', 'first_name': 'Sophie', 'last_name': 'FONFEC',
            'password': 'test2pwd', 'email': 'test2@chrysal.io',
            'attachments_key': 'User2'})
        self.add_user({
            'login': 'test3', 'first_name': 'Guy', 'last_name': 'LIGUILI',
            'password': 'test3pwd', 'email': 'test3@chrysal.io',
            'attachments_key': 'User3'})

        self.configurator.add_route('home', '/')
        self.configurator.add_route('user_index', '/user/index')
        self.configurator.add_route('user_create', '/user/create')
        self.configurator.add_route('user_edit', '/user/edit/{user_id}')
        self.configurator.add_route('user_view', '/user/view/{user_id}')
        self.configurator.add_route(
            'user_password_reset', '/user/password/reset/{user_id}/{token}')

    # -------------------------------------------------------------------------
    def test_index(self):
        """[u:views.user.UserView.index]"""
        from collections import namedtuple
        from ..lib.filter import Filter
        from ..lib.paging import Paging
        from ..views.user import UserView
        from . import DummyPOST

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_index')

        # Normal HTTP
        response = UserView(self.request).index()
        self.assertIn('pfilter', response)
        self.assertIn('paging', response)
        self.assertIn('i_creator', response)
        self.assertIn('i_editor', response)
        self.assertIsInstance(response['pfilter'], Filter)
        self.assertFalse(response['pfilter'].is_empty())
        self.assertIsInstance(response['paging'], Paging)
        self.assertTrue(response['i_creator'])
        self.assertTrue(response['i_editor'])
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Users')

        # Ajax
        self.request.is_xhr = True
        self.request.POST = DummyPOST()
        response = UserView(self.request).index()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)

    # -------------------------------------------------------------------------
    def test_index_send_invitation(self):
        """[u:views.user.UserView.index] send invitation"""
        from ..views.user import UserView

        self.request.registry['settings']['title'] = 'Chrysalio'
        self.request.registry.settings['smtp.host'] = 'foo'
        self.request.session['user'] = {
            'user_id': 1, 'login': 'test1', 'email': 'test1@chrysal.io'}
        self.request.POST = {'mel!1.x': True}

        UserView(self.request).index()
        self.assertTrue(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_index_delete(self):
        """[u:views.user.UserView.index] delete user"""
        from os import makedirs
        from os.path import join, exists
        from . import TEST_DIR
        from ..views.user import UserView
        from ..models.dbuser import DBUser

        dbsession = self.request.dbsession
        self.assertEqual(dbsession.query(DBUser).count(), 4)
        self.request.session['user'] = {'user_id': 1}
        attachments = join(TEST_DIR, 'Attachments')
        self.request.registry.settings['attachments'] = attachments

        self.request.POST = {'del!1.x': True}
        UserView(self.request).index()
        self.assertEqual(dbsession.query(DBUser).count(), 4)

        self.request.session['user'] = {'user_id': 2}
        UserView(self.request).index()
        self.assertEqual(dbsession.query(DBUser).count(), 4)

        makedirs(join(attachments, 'Users', 'User2'))
        self.request.POST = {'del!3.x': True}
        UserView(self.request).index()
        self.assertEqual(dbsession.query(DBUser).count(), 3)
        self.assertFalse(exists(join(attachments, 'Users', 'User2')))

    # -------------------------------------------------------------------------
    def test_index_import(self):
        """[u:views.user.UserView.index] import users"""
        from os.path import join, exists
        from cgi import FieldStorage
        from . import TEST_DIR, TEST1_USR_ZIP, TEST3_USR_XML, DummyPOST
        from ..models.dbuser import DBUser
        from ..views.user import UserView

        self.request.POST = DummyPOST({'imp!.x': True})
        response = UserView(self.request).index()
        self.assertIn('paging', response)
        self.assertNotIn('user1', [k.login for k in response['paging']])

        self.request.registry.settings['attachments'] = TEST_DIR
        with open(TEST1_USR_ZIP, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_USR_ZIP
            input_file.file = hdl
            self.request.POST.multikeys = {'file': (input_file,)}
            response = UserView(self.request).index()
        self.assertIn('paging', response)
        self.assertIn('user3', [k.login for k in response['paging']])
        self.assertTrue(exists(join(
            TEST_DIR, DBUser.attachments_dir, 'User3', 'user3.svg')))

        with open(TEST3_USR_XML, 'r') as hdl:
            input_file = FieldStorage()
            input_file.file = hdl
            input_file.filename = TEST3_USR_XML
            self.request.POST.multikeys = {'file': (input_file,)}
            response = UserView(self.request).index()
        self.assertIn('paging', response)
        self.assertIn('user1', [k.login for k in response['paging']])

    # -------------------------------------------------------------------------
    def test_index_export(self):
        """[u:views.user.UserView.index] export users"""
        from pyramid.response import Response
        from ..views.user import UserView

        self.request.POST = {'exp!1.x': True}
        UserView(self.request).index()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('cannot export', errors[0])

        self.request.POST = {'exp!2.x': True}
        response = UserView(self.request).index()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'test1', response.body)

    # -------------------------------------------------------------------------
    def test_index_filter(self):
        """[u:views.user.UserView.index_filter]"""
        from ..views.user import UserView

        self.request.is_xhr = True
        self.request.params = {'field': 'login', 'value': 'test'}
        data = UserView(self.request).index_filter()
        self.assertIsInstance(data, list)
        self.assertIn('test1', data)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:views.user.UserView.view]"""
        from collections import namedtuple
        from pyramid.httpexceptions import HTTPNotFound, HTTPForbidden
        from ..views.user import UserView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_view')
        self.request.session['user'] = {'user_id': 1}

        self.request.matchdict = {'user_id': 'foo'}
        self.assertRaises(HTTPNotFound, UserView(self.request).view)

        self.request.matchdict = {'user_id': '11'}
        self.assertRaises(HTTPNotFound, UserView(self.request).view)

        self.request.matchdict = {'user_id': '2'}
        self.request.referrer = '/user/view/1'
        response = UserView(self.request).view()
        self.assertIn('form', response)
        self.assertIn('user_name', response)
        self.assertEqual(response['user_name'], 'Édith AVULEUR')
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Édith AVULEUR Account')
        self.assertEqual(
            self.request.breadcrumbs.current_path(), '/user/view/2')

        self.request.has_permission = lambda x: False
        self.request.session['user'] = {'user_id': 3}
        self.assertRaises(HTTPForbidden, UserView(self.request).view)

    # -------------------------------------------------------------------------
    def test_view_export(self):
        """[u:views.user.UserView.view] export users"""
        from collections import namedtuple
        from pyramid.response import Response
        from ..views.user import UserView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_view')
        self.request.POST = {'exp!.x': True}
        self.request.matchdict = {'user_id': '2'}
        self.request.session['user'] = {'user_id': 1}

        response = UserView(self.request).view()
        self.assertIsInstance(response, Response)
        self.assertEqual(response.content_type, 'application/xml')
        self.assertIn(b'test1', response.body)

    # -------------------------------------------------------------------------
    def test_view_send_invitation(self):
        """[u:views.user.UserView.view] send invitation"""
        from collections import namedtuple
        from ..views.user import UserView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_view')
        self.request.registry['settings']['title'] = 'Chrysalio'
        self.request.registry.settings['smtp.host'] = 'foo'
        self.request.session['user'] = {
            'user_id': 2, 'login': 'test2', 'email': 'test2@chrysal.io'}
        self.request.matchdict = {'user_id': '2'}
        self.request.POST = {'mel!.x': True}

        UserView(self.request).view()
        self.assertTrue(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_create(self):
        """[u:views.user.UserView.edit] creation"""
        from collections import namedtuple
        from pyramid.httpexceptions import HTTPForbidden
        from ..views.user import UserView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_create')

        response = UserView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('user_name', response)
        self.assertIsNone(response['user_name'])
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'User Creation')

        self.request.has_permission = lambda x: False
        self.assertRaises(HTTPForbidden, UserView(self.request).edit)

    # -------------------------------------------------------------------------
    def test_create_save(self):
        """[u:views.user.UserView.edit] save creation"""
        from collections import namedtuple
        from json import dumps
        from ..models.dbprofile import DBProfile
        from ..views.user import UserView

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_viewer', i18n_label=dumps({'en': 'User Viewer'})))
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_create')
        self.request.POST = {
            'login': 'test1', 'status': 'active',
            'first_name': 'Sébastienne', 'last_name': 'TOUSSEUL',
            'email': 'test11@chrysal.io',
            'password1': 'test11pwd', 'password2': 'test11pwd',
            'email_hidden': True, 'pfl:user_viewer': True, 'sav!.x': True}

        # Existing user
        response = UserView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertIn('form', response)
        self.assertEqual(len(errors), 1)
        self.assertIn('exists', errors[0])
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'User Creation')
        dbsession.rollback()

        # New user, missing status
        del self.request.POST['status']
        UserView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])

        # New user
        self.request.POST['login'] = 'test11'
        self.request.POST['status'] = 'active'
        UserView(self.request).edit()
        self.assertFalse(self.request.session.pop_flash('alert'))

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[u:views.user.UserView.edit]"""
        from collections import namedtuple
        from pyramid.httpexceptions import HTTPForbidden
        from ..models.dbuser import DBUser
        from ..views.user import UserView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_edit')
        self.request.matchdict = {'user_id': '2'}
        self.request.session['user'] = {'user_id': 2}

        # As an administrator
        self.request.referrer = '/user/account'
        response = UserView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('user_name', response)
        self.assertEqual(response['user_name'], 'Édith AVULEUR')
        self.assertIn('dbuser', response)
        self.assertIsInstance(response['dbuser'], DBUser)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'My Account Edition')

        # As a normal user
        self.request.has_permission = lambda x: False
        self.request.session['user'] = {'user_id': 3}
        self.assertRaises(HTTPForbidden, UserView(self.request).edit)

    # -------------------------------------------------------------------------
    def test_edit_save(self):
        """[u:views.user.UserView.edit] save"""
        from json import dumps
        from pyramid.httpexceptions import HTTPFound
        from ..models.dbprofile import DBProfile
        from ..models.dbuser import DBUser, DBUserProfile
        from ..models.dbgroup import DBGroupUser
        from ..views.user import UserView

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='user_editor', i18n_label=dumps({'en': 'User Editor'})))
        dbprofile = DBProfile(
            profile_id='user_viewer', i18n_label=dumps({'en': 'User Viewer'}))
        dbsession.add(dbprofile)
        dbuser = dbsession.query(DBUser).filter_by(login='test1').first()
        dbuser.profiles.append(dbprofile)
        dbuser.groups.append(DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'})))
        dbsession.add(DBGroup(
            group_id='editors', i18n_label=dumps({'en': 'Editors'})))
        self.request.matchdict = {'user_id': str(dbuser.user_id)}
        self.request.session['user'] = {'user_id': dbuser.user_id}

        self.request.POST = {
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'password1': 'test1pwd2', 'password2': 'test1pwd2',
            'email': 'test1chrysal.io', 'pfl:user_editor': True,
            'grp:editors': True, 'sav!.x': True}
        UserView(self.request).edit()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('Correct', errors[0])

        self.request.POST['email'] = 'test1@chrysal.io'
        self.assertIsInstance(UserView(self.request).edit(), HTTPFound)
        self.assertIsNone(dbsession.query(DBUserProfile).filter_by(
            user_id=dbuser.user_id, profile_id='user_viewer').first())
        self.assertIsNotNone(dbsession.query(DBUserProfile).filter_by(
            user_id=dbuser.user_id, profile_id='user_editor').first())
        self.assertIsNone(dbsession.query(DBGroupUser).filter_by(
            group_id='managers', user_id=dbuser.user_id).first())
        self.assertIsNotNone(dbsession.query(DBGroupUser).filter_by(
            group_id='editors', user_id=dbuser.user_id).first())

        self.request.has_permission = lambda x: False
        self.request.POST['grp:managers'] = True
        UserView(self.request).edit()
        self.assertIsNone(dbsession.query(DBGroupUser).filter_by(
            group_id='managers', user_id=dbuser.user_id).first())

    # -------------------------------------------------------------------------
    def test_edit_picture(self):
        """[u:views.user.UserView.edit] change picture"""
        from ..views.user import UserView

        self.request.matchdict = {'user_id': '2'}
        self.request.session['user'] = {'user_id': 2}
        self.request.POST = {
            'login': 'test1', 'first_name': 'Édith', 'last_name': 'AVULEUR',
            'email': 'test1@chrysal.io', 'picture': None, 'pct!.x': True}

        # Normal HTTP
        UserView(self.request).edit()

        # AJAX
        self.request.is_xhr = True
        response = UserView(self.request).edit()
        self.assertIsInstance(response, dict)
        self.assertFalse(response)

    # -------------------------------------------------------------------------
    def test_password_forgot(self):
        """[u:views.user.UserView.password_forgot]"""
        from collections import namedtuple
        from ..views.user import UserView

        self.request.registry['settings']['title'] = 'Chrysalio'
        self.request.registry['settings']['email'] = 'admin@chrysal.io'
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_password_forgot')

        response = UserView(self.request).password_forgot()
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Forgot Password')

        self.request.POST = {'email': 'testX@chrysal.io'}
        UserView(self.request).password_forgot()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertIn('email associated', errors[0])

        self.request.registry.settings['smtp.host'] = 'foo'
        self.request.POST['email'] = 'test1@chrysal.io'
        UserView(self.request).password_forgot()
        errors = self.request.session.pop_flash('alert')
        self.assertEqual(len(errors), 1)
        self.assertTrue(
            ('Connection refused' in errors[0]) or
            ('No address associated with hostname' in errors[0]) or
            ('Name or service not known' in errors[0]) or
            ('Temporary failure in name resolution') in errors[0] or
            ('Connection timed out' in errors[0]))

    # -------------------------------------------------------------------------
    def test_password_reset(self):
        """[u:views.user.UserView.password_reset]"""
        from hashlib import sha1
        from datetime import date
        from collections import namedtuple
        from pyramid.httpexceptions import HTTPNotFound
        from ..views.user import UserView

        self.request.registry['settings']['title'] = 'Chrysalio'
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_password_reset')
        self.request.matchdict['user_id'] = '2'
        self.request.matchdict['token'] = 'foo'

        self.assertRaises(HTTPNotFound, UserView(self.request).password_reset)

        token = 'test1{0}'.format(date.today().isoformat())
        token = sha1(token.encode('utf8')).hexdigest()
        self.request.matchdict['token'] = token
        response = UserView(self.request).password_reset()
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Password Reset')

        self.request.POST = {
            'password1': 'test1pwd', 'password2': 'test1pwd'}
        UserView(self.request).password_reset()
        messages = self.request.session.pop_flash()
        self.assertEqual(len(messages), 1)
        self.assertIn('changed', messages[0])

    # -------------------------------------------------------------------------
    def test_can_view(self):
        """[u:views.user.UserView.can_view]"""
        from ..views.user import UserView

        self.assertTrue(UserView(self.request).can_view())

        self.request.has_permission = lambda x: False
        self.request.session['user'] = {'user_id': 1}
        self.request.matchdict['user_id'] = '1'
        self.assertTrue(UserView(self.request).can_view())

        self.request.matchdict['user_id'] = '2'
        self.assertFalse(UserView(self.request).can_view())

    # -------------------------------------------------------------------------
    def test_can_create(self):
        """[u:views.user.UserView.can_create]"""
        from ..views.user import UserView

        self.assertTrue(UserView(self.request).can_create())

        self.request.has_permission = lambda x: False
        self.assertFalse(UserView(self.request).can_create())

    # -------------------------------------------------------------------------
    def test_can_edit(self):
        """[u:views.user.UserView.can_edit]"""
        from ..views.user import UserView

        self.assertTrue(UserView(self.request).can_edit())

        self.request.has_permission = lambda x: False
        self.request.session['user'] = {'user_id': 1}
        self.request.matchdict['user_id'] = '1'
        self.assertTrue(UserView(self.request).can_edit())

        self.request.matchdict['user_id'] = '2'
        self.assertFalse(UserView(self.request).can_edit())


# =============================================================================
class FViewsUserUserView(FunctionalTestCase):
    """Functional test class for testing :class:`views.user.UserView`."""

    # -------------------------------------------------------------------------
    def test_index(self):
        """[f:views.user.UserView.index]"""
        self.login('test1')
        self.testapp.get('/user/index', status=403)

        self.login('admin')
        response = self.testapp.get('/user/index?display=list', status=200)
        self.assertIn(b'Users', response.body)
        self.assertIn(b'Create', response.body)
        self.assertIn(b'Import', response.body)
        self.assertIn(b'Send an invitation', response.body)
        self.assertIn(b'Export user', response.body)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[f:views.user.UserView.view]"""
        self.login('test1')
        self.testapp.get('/user/view/3', status=403)

        response = self.testapp.get('/user/view/2', status=200)
        self.assertIn(b'Edit', response.body)

    # -------------------------------------------------------------------------
    def test_create(self):
        """[f:views.user.UserView.create]"""
        self.login('test1')
        self.testapp.get('/user/create', status=403)

        self.login('admin')
        response = self.testapp.get('/user/create', status=200)
        self.assertIn(b'Create', response.body)

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[f:views.user.UserView.edit]"""
        self.login('test1')
        self.testapp.get('/user/edit/3', status=403)

        response = self.testapp.get('/user/edit/2', status=200)
        self.assertIn(b'Save', response.body)

    # -------------------------------------------------------------------------
    # def test_password_forgot(self):
    #     """[f:views.user.UserView.password_forgot]"""
    #     csrf = self.testapp.get('/login').form.get('csrf_token').value
    #     self.testapp.post(
    #         '/user/password/forgot',
    #         {'email': 'test1@chrysal.io', 'csrf_token': csrf}, status=302)

    # -------------------------------------------------------------------------
    def test_password_reset(self):
        """[f:views.user.UserView.password_reset]"""
        from hashlib import sha1
        from datetime import date

        csrf = self.testapp.get('/login').form.get('csrf_token').value
        token = 'test1{0}'.format(date.today().isoformat())
        token = sha1(token.encode('utf8')).hexdigest()
        self.testapp.post(
            '/user/password/reset/2/{0}'.format(token), {
                'password1': 'test1pwd2', 'password2': 'test1pwd2',
                'csrf_token': csrf},
            status=302)



# pylint: disable = import-outside-toplevel
"""Tests of ``views.home`` functions."""

from pyramid.httpexceptions import HTTPFound
from pyramid.testing import DummySession

from . import DBUnitTestCase, FunctionalTestCase


# =============================================================================
class UViewsHomeHomeView(DBUnitTestCase):
    """Unit test class for testing :func:`views.home.home_view`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:views.home.home_view]"""
        from collections import namedtuple
        from ..views.home import home_view
        from ..modes import MODE_HOME, MODE_ADMIN

        self.configurator.registry['modes'] = [MODE_HOME, MODE_ADMIN]
        self.configurator.add_route('home', '/')
        self.configurator.add_route('mode', '/mode/{mode_id}')
        self.configurator.add_route('settings_view', '/settings/view')
        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        self.request.session = DummySession({'modes': ['home', 'Home', None]})

        # Display home
        response = home_view(self.request)
        self.assertIn('button', response)
        self.assertEqual(self.request.breadcrumbs.current_title(), 'Home')

        # With a mode and a page change
        self.request.params = {'route': 'settings_view', 'mode': 'admin'}
        response = home_view(self.request)
        self.assertIsInstance(response, HTTPFound)


# =============================================================================
class FViewsHomeHomeView(FunctionalTestCase):
    """Functional test class for testing :func:`views.home.home_view`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[f:views.home.home_view]"""
        self.testapp.get('/', status=302)
        self.login('test1')
        self.testapp.get('/', status=200)



# -*- coding: utf-8 -*-
"""SQLAlchemy-powered model definitions for user groups."""

from json import dumps

import colander
from sqlalchemy import Column, ForeignKey, Integer, String, Text, PickleType
from sqlalchemy.orm import relationship
from lxml import etree
from webhelpers2.html import literal

from ..lib.i18n import _, record_format_i18n
from ..lib.i18n import schema_i18n_labels, defaults_i18n_labels
from ..lib.i18n import view_i18n_labels, edit_i18n_labels
from ..lib.utils import make_id
from ..lib.xml import i18n_xml_text, db2xml_i18n_labels
from . import DBDeclarativeClass, ID_LEN, LABEL_LEN, DESCRIPTION_LEN
from .dbbase import DBBaseClass
from .dbuser import DBUser
from .dbprofile import DBProfile


# =============================================================================
class DBGroup(DBDeclarativeClass, DBBaseClass):
    """SQLAlchemy-powered user group class."""

    suffix = 'ciogrp'
    attachments_dir = 'Groups'
    _settings_tabs = (_('Information'), _('Users'), _('Profiles'))

    __tablename__ = 'groups'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    group_id = Column(String(ID_LEN), primary_key=True)
    i18n_label = Column(Text(), nullable=False)
    i18n_description = Column(PickleType(1))
    attachments_key = Column(String(ID_LEN + 20))
    picture = Column(String(ID_LEN + 4))

    users = relationship('DBGroupUser', cascade='all, delete')
    profiles = relationship(DBProfile, secondary='groups_profiles')

    # -------------------------------------------------------------------------
    @classmethod
    def xml2db(cls, dbsession, group_elt, error_if_exists=True, kwargs=None):
        """Load a user group from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  group_elt: lxml.etree.Element
        :param group_elt:
            User group XML element.
        :param bool error_if_exists: (default=True)
            It returns an error if user group already exists.
        :param dict kwargs: (optional)
            Dictionary of keyword arguments.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        # pylint: disable = unused-argument
        # Check if already exists
        group_id = make_id(group_elt.get('id'), 'token', ID_LEN)
        dbgroup = dbsession.query(cls).filter_by(
            group_id=group_id).first()
        if dbgroup is not None:
            if error_if_exists:
                return _('Group "${g}" already exists.', {'g': group_id})
            return None

        # Create group
        record = cls.record_from_xml(group_id, group_elt)
        error = cls.record_format(record)
        if error:
            return error
        dbgroup = cls(**record)
        dbsession.add(dbgroup)

        return dbgroup.xml2db_extra(dbsession, group_elt, kwargs)

    # -------------------------------------------------------------------------
    def xml2db_extra(self, dbsession, group_elt, kwargs):
        """Load extra information on a group from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  group_elt: lxml.etree.Element
        :param group_elt:
            Group XML element.
        :param dict kawrgs:
            Dictionary of keyword arguments with the key ``'profiles'``.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        dbsession.flush()

        # Add users
        refs = kwargs.get('users', {})
        done = set()
        for elt in group_elt.xpath('users/user'):
            user_id = refs.get(elt.text)
            if user_id is not None and user_id not in done:
                dbsession.add(DBGroupUser(
                    group_id=self.group_id, user_id=user_id))
                done.add(user_id)

        # Add profiles
        profile_ids = kwargs.get('profiles') if kwargs else None
        if profile_ids:
            done = set()
            for elt in group_elt.findall('profiles/profile'):
                profile_id = elt.text
                if profile_id not in done and profile_id in profile_ids:
                    dbsession.add(DBGroupProfile(
                        group_id=self.group_id, profile_id=profile_id))
                    done.add(profile_id)

    # -------------------------------------------------------------------------
    @classmethod
    def record_from_xml(cls, group_id, group_elt):
        """Convert an user group XML element into a dictionary.

        :param str group_id:
            User group ID.
        :type  group_elt: lxml.etree.Element
        :param group_elt:
            Group XML element.
        :rtype: dict
        """
        attachments_elt = group_elt.find('attachments')
        return {
            'group_id': group_id,
            'i18n_label': dumps(
                i18n_xml_text(group_elt, 'label'), ensure_ascii=False),
            'i18n_description': i18n_xml_text(group_elt, 'description'),
            'attachments_key':
            attachments_elt is not None and attachments_elt.get('key') or None,
            'picture':
            attachments_elt is not None and attachments_elt.findtext(
                'picture') or None}

    # -------------------------------------------------------------------------
    @classmethod
    def record_format(cls, record):
        """Check and possibly correct a record before inserting it in the
        database.

        :param dict record:
            Dictionary of values to check.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        for k in [i for i in record if record[i] is None]:
            del record[k]

        # Group ID
        if not record.get('group_id'):
            return _('User group without ID.')
        record['group_id'] = make_id(record['group_id'], 'xmlid', ID_LEN)

        # Labels and descriptions
        if not record_format_i18n(record):
            return _('User group without label.')
        if not record.get('i18n_description'):
            record['i18n_description'] = {}
        return None

    # -------------------------------------------------------------------------
    def db2xml(self, dbsession):
        """Serialize an user group to a XML representation.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :rtype: lxml.etree.Element
        """
        group_elt = etree.Element('group')
        group_elt.set('id', self.group_id)

        # Labels and descriptions
        db2xml_i18n_labels(self, group_elt, 3)

        # Attachments
        if self.attachments_key:
            elt = etree.SubElement(
                group_elt, 'attachments', key=self.attachments_key)
            if self.picture:
                etree.SubElement(elt, 'picture').text = self.picture

        # Users
        if self.users:
            users = {}
            users_elt = etree.SubElement(group_elt, 'users')
            for dbuser in self.users:
                if dbuser.user_id not in users:
                    users[dbuser.user_id] = dbsession.query(
                        DBUser.login).filter_by(
                            user_id=dbuser.user_id).first()[0]
                etree.SubElement(users_elt, 'user').text = \
                    users[dbuser.user_id]

        # Profiles
        if self.profiles:
            elt = etree.SubElement(group_elt, 'profiles')
            for item in self.profiles:
                etree.SubElement(elt, 'profile').text = item.profile_id

        return group_elt

    # -------------------------------------------------------------------------
    def tab4view(self, request, tab_index, form, user_filter, user_paging):
        """Generate the tab content of a user group.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  user_filter: chrysalio.lib.filter.Filter
        :param user_filter:
            Filter for users.
        :type  user_paging: chrysalio.lib.paging.Paging
        :param user_paging:
            Paging for warehouse users.
        :rtype: webhelpers2.html.literal
        """
        if tab_index == 0:
            return self._tab4view_information(request, form)
        if tab_index == 1:
            return self._tab4view_users(
                request, form, user_filter, user_paging)
        if tab_index == 2:
            return self._tab4view_profiles(request)
        return ''

    # -------------------------------------------------------------------------
    def _tab4view_information(self, request, form):
        """Generate the information tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = form.grid_item(
            translate(_('Identifier:')), self.group_id, clear=True)
        html += view_i18n_labels(request, form, self)
        return html

    # -------------------------------------------------------------------------
    def _tab4view_users(self, request, form, user_filter, user_paging):
        """Generate the users tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  user_filter: chrysalio.lib.filter.Filter
        :param user_filter:
            Filter for users.
        :type  user_paging: chrysalio.lib.paging.Paging
        :param user_paging:
            Paging for warehouse users.
        :rtype: webhelpers2.html.literal
        """
        html = DBUser.paging_filter(
            request, form, user_filter, user_paging)
        html += self._user_thead(request, user_paging)
        for dbuser in user_paging:
            html += '<tr>'\
                '<th><a href="{user_view}">{login}</a></th>'\
                '<td class="cioOptional">{fname}</td><td>{lname}</td>'\
                '</tr>\n'.format(
                    user_view=request.route_path(
                        'user_view', user_id=dbuser.user_id),
                    login=dbuser.login,
                    fname=dbuser.first_name or '', lname=dbuser.last_name)
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    def _tab4view_profiles(self, request):
        """Generate the profiles tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if not self.profiles:
            return translate(_('No attributed profile.'))

        html = '<table>\n<thead>\n'\
            '<tr><th>{label}</th><th>{description}</th></tr>\n'\
            '</thead>\n<tbody>\n'.format(
                label=translate(_('Label')),
                description=translate(_('Description')))
        for dbprofile in sorted(self.profiles, key=lambda k: k.profile_id):
            html += \
                '<tr><th><a href="{profile_view}">{label}</a></th>'\
                '<td>{description}</td></tr>\n'.format(
                    profile_view=request.route_path(
                        'profile_view', profile_id=dbprofile.profile_id),
                    label=dbprofile.label(request),
                    description=dbprofile.description(request))
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def settings_schema(cls, request, defaults, profiles, dbgroup=None):
        """Return a Colander schema to edit user group.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param dict defaults:
            Default values for the form set by the user paging object.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :type  dbgroup: DBGroup
        :param dbgroup: (optional)
            Current user group SqlAlchemy object.
        :rtype: tuple
        :return:
            A tuple such as ``(schema, defaults)``.
        """
        # Schema
        schema = colander.SchemaNode(colander.Mapping())
        if dbgroup is None:
            schema.add(colander.SchemaNode(
                colander.String(), name='group_id',
                validator=colander.All(
                    colander.Regex(r'^[a-z][a-z0-9_]+$'),
                    colander.Length(min=2, max=ID_LEN))))

        schema_i18n_labels(request, schema, LABEL_LEN, DESCRIPTION_LEN)

        # Profiles
        if request.has_permission('group-create'):
            for profile_id in profiles:
                schema.add(colander.SchemaNode(
                    colander.Boolean(), name='pfl:{0}'.format(profile_id),
                    missing=False))

        # Defaults
        if dbgroup is not None:
            defaults.update(defaults_i18n_labels(dbgroup))
            for dbitem in dbgroup.users:
                defaults['mbr:{0}'.format(dbitem.user_id)] = True
            for dbitem in dbgroup.profiles:
                defaults['pfl:{0}'.format(dbitem.profile_id)] = True

        return schema, defaults

    # -------------------------------------------------------------------------
    @classmethod
    def tab4edit(cls, request, tab_index, form, user_filter, user_paging,
                 profiles, dbgroup=None):
        """Generate the tab content of user group for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int tab_index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  user_filter: chrysalio.lib.filter.Filter
        :param user_filter:
            Filter for users.
        :type  user_paging: chrysalio.lib.paging.Paging
        :param user_paging:
            Paging for all users.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :type  dbgroup: DBGroup
        :param dbgroup: (optional)
            Current user group SqlAlchemy object.
        :rtype: webhelpers2.html.literal
        """
        # pylint: disable = too-many-arguments
        if tab_index == 0:
            return cls._tab4edit_information(request, form, dbgroup)
        if tab_index == 1:
            return cls._tab4edit_users(
                request, form, user_filter, user_paging)
        if tab_index == 2:
            return cls._tab4edit_profiles(request, form, profiles)
        return ''

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_information(cls, request, form, dbgroup):
        """Generate the information tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  dbgroup: DBGroup
        :param dbgroup:
            Current user group SqlAlchemy object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if dbgroup is None:
            html = form.grid_text(
                'group_id', translate(_('Identifier:')), required=True,
                maxlength=ID_LEN, clear=True)
        else:
            html = form.grid_item(
                translate(_('Identifier:')), dbgroup.group_id, clear=True)
        html += edit_i18n_labels(request, form, LABEL_LEN, DESCRIPTION_LEN)
        return html

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_users(cls, request, form, user_filter, user_paging):
        """Generate the users tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  user_filter: chrysalio.lib.filter.Filter
        :param user_filter:
            Filter for users.
        :type  user_paging: chrysalio.lib.paging.Paging
        :param user_paging:
            Paging for all users.
        :rtype: webhelpers2.html.literal
        """
        html = DBUser.paging_filter(
            request, form, user_filter, user_paging)
        html += cls._user_thead(request, user_paging).replace(
            '<tr>', '<tr><th class="cioCheckbox" id="check_all"></th>')
        for dbuser in user_paging:
            html += '<tr>'\
                '<td class="cioCheckbox cioSelect">{check}{shown}</td>'\
                '<th>{login}</th>'\
                '<td class="cioOptional">{fname}</td><td>{lname}</td>'\
                '</tr>\n'.format(
                    check=form.custom_checkbox(
                        'mbr:{0}'.format(dbuser.user_id)),
                    shown=form.hidden('shw:{0}'.format(dbuser.user_id)),
                    login=dbuser.login,
                    fname=dbuser.first_name or '', lname=dbuser.last_name)
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_profiles(cls, request, form, profiles):
        """Generate the profiles tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :param dict profiles:
            A dictionary such as ``{profile_id: (label, description),...}``.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if not profiles:
            return translate(_('No available profile.'))

        if not request.has_permission('user-create'):
            return translate(
                _('You do not have the rigths to edit profiles.'))

        html = '<table>\n<thead>\n'\
            '<tr><th></th><th>{label}</th>'\
            '<th>{description}</th></tr>\n</thead>\n<tbody>\n'.format(
                label=translate(_('Label')),
                description=translate(_('Description')))
        for profile_id in sorted(profiles):
            html += \
                '<tr><td>{selected}</td>'\
                '<th><label for="{id}">{label}</label></th>'\
                '<td>{description}</td></tr>\n'.format(
                    selected=form.custom_checkbox(
                        'pfl:{0}'.format(profile_id)),
                    id='pfl{0}'.format(profile_id),
                    label=profiles[profile_id][0],
                    description=profiles[profile_id][1])
        html += '</tbody>\n</table>\n'

        return literal(html)

    # -------------------------------------------------------------------------
    @classmethod
    def _user_thead(cls, request, user_paging):
        """Table header for group users.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  user_paging: chrysalio.lib.paging.Paging
        :param user_paging:
            Paging for users.
        :rtype: str
        """
        translate = request.localizer.translate
        return \
            '<table class="cioPagingList">\n<thead><tr>'\
            '<th>{login}</th>'\
            '<th class="cioOptional">{fname}</th><th>{lname}</th>'\
            '</tr></thead>\n<tbody>\n'.format(
                login=user_paging.sortable_column(
                    translate(_('Login')), 'login'),
                fname=user_paging.sortable_column(
                    translate(_('First name')), 'first_name'),
                lname=user_paging.sortable_column(
                    translate(_('Last name')), 'last_name'))


# =============================================================================
class DBGroupUser(DBDeclarativeClass):
    """Class to link groups with their authorized users (many-to-many)."""
    # pylint: disable = too-few-public-methods

    __tablename__ = 'groups_users'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    group_id = Column(
        String(ID_LEN), ForeignKey('groups.group_id', ondelete='CASCADE'),
        primary_key=True)
    user_id = Column(
        Integer, ForeignKey('users.user_id', ondelete='CASCADE'),
        primary_key=True)


# =============================================================================
class DBGroupProfile(DBDeclarativeClass):
    """Class to link groups with their profiles (many-to-many)."""
    # pylint: disable = too-few-public-methods

    __tablename__ = 'groups_profiles'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    group_id = Column(
        String(ID_LEN), ForeignKey('groups.group_id', ondelete='CASCADE'),
        primary_key=True)
    profile_id = Column(
        String(ID_LEN), ForeignKey('profiles.profile_id', ondelete='CASCADE'),
        primary_key=True)



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.restful`` functions."""

from datetime import datetime, timedelta
from unittest import TestCase

from . import DBUnitTestCase


# =============================================================================
class ULibRestfullRestfulCall(TestCase):
    """Unit test class for :func:`lib.restful.restful_call`."""

    # -------------------------------------------------------------------------
    def test_it_ko(self):
        """[u:lib.restful.restful_call] Ko"""
        from ..lib.utils import decrypt
        from ..lib.restful import restful_call
        from . import RESTFUL_HOST, RESTFUL_LOGIN, RESTFUL_KEY

        # Incorrect URL
        url = 'http://localhost:6543/user/index'
        response, error = restful_call(url, RESTFUL_LOGIN, 'foo')
        self.assertIsNone(response)
        self.assertTrue(
            ('Connection refused' in error) or
            ('Connection timed out' in error) or
            ('No route to host') in error or
            ('Could not resolve host') in error or
            ('HTTP Error 400: Bad Request') in error)

        # Incorrect action
        url = '{0}warehouse/fullrefresh/Buggy'.format(RESTFUL_HOST)
        response, error = restful_call(
            url, RESTFUL_LOGIN, decrypt(RESTFUL_KEY, 'test'))
        self.assertIsNone(response)
        self.assertIsNotNone(error)

    # -------------------------------------------------------------------------
    def test_it_ok(self):
        """[u:lib.restful.restful_call] Ok"""
        from ..lib.utils import decrypt
        from ..lib.restful import restful_call
        from . import RESTFUL_HOST, RESTFUL_LOGIN, RESTFUL_KEY

        # Without data
        url = '{0}warehouse/fullrefresh/Sandbox'.format(RESTFUL_HOST)
        response, error = restful_call(
            url, RESTFUL_LOGIN, decrypt(RESTFUL_KEY, 'test'))
        if response is None:
            return
        self.assertIsNotNone(response)
        self.assertIsNone(error)

        # With data
        url = '{0}warehouse/refresh/Sandbox'.format(RESTFUL_HOST)
        data = {
            'files': ['file_00001.txt', 'file_00002.txt'],
            'recursive': True, 'ignored': {'foo': 1}}
        response, error = restful_call(
            url, RESTFUL_LOGIN, decrypt(RESTFUL_KEY, 'test'), data)
        self.assertIsNotNone(response)
        self.assertIsNone(error)


# =============================================================================
class ULibRestfullRestfulLogin(DBUnitTestCase):
    """Unit test class for :func:`lib.restful.restful_login`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.restful.restful_login]"""
        from ..lib.utils import decrypt
        from ..lib.restful import restful_login
        from ..lib.utils import encrypt
        from . import RESTFUL_KEY

        # User not in session, incorrect login
        key = decrypt(RESTFUL_KEY, 'test')
        self.assertIsNotNone(restful_login(self.request, key))

        # User not in session, correct login, no token
        self.add_user({
            'login': 'admin', 'status': 'administrator',
            'last_name': 'Administrator', 'password': 'adminpwd',
            'email': 'admin@chrysal.io'})
        self.request.params = {'login': 'admin'}
        self.assertIsNotNone(restful_login(self.request, key))

        # User not in session, correct login, bad token
        self.request.params['token'] = 'bad-token'
        self.assertIsNotNone(restful_login(self.request, key))

        # User not in session, correct login, expired token
        self.request.params['token'] = encrypt(
            (datetime.utcnow() + timedelta(hours=1)).isoformat(), key)
        self.assertIsNotNone(restful_login(self.request, key))

        # User not in session, correct login, correct token, incorrect user
        self.request.params['login'] = 'foo'
        self.request.params['token'] = encrypt(
            datetime.utcnow().isoformat(), key)
        self.assertIsNotNone(restful_login(self.request, key))

        # User not in session, correct login, correct token, correct user
        self.request.params['login'] = 'admin'
        self.request.params['token'] = encrypt(
            datetime.utcnow().isoformat(), key)
        self.assertIsNone(restful_login(self.request, key))

        # User in session
        self.assertIsNone(restful_login(self.request, key))



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.config`` classes and functions."""

from os.path import dirname
from configparser import ConfigParser
from unittest import TestCase

from pyramid.security import Authenticated, Allow, ALL_PERMISSIONS

from . import ConfiguratorTestCase


# =============================================================================
class ConfigTestCase(TestCase):
    """Base class for testing config functions."""

    config = None

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..lib.utils import tounicode
        from . import TEST_INI

        self.config = ConfigParser({'here': dirname(TEST_INI)})
        self.config.read(tounicode(TEST_INI), encoding='utf8')


# =============================================================================
class ULibConfigUpdateACL(ConfiguratorTestCase):
    """Unit test class for :func:`lib.config.update_acl`."""

    # -------------------------------------------------------------------------
    def test_update(self):
        """[u:lib.config.update_acl]"""
        from ..lib.config import update_acl

        class DummyRootFactory(object):
            """Dummy Access Control List (ACL) definition."""
            # pylint: disable = too-few-public-methods
            __acl__ = [
                (Allow, Authenticated, 'authenticated'),
                (Allow, 'system.administrator', ALL_PERMISSIONS)]

            def __init__(self, request):
                """Constructor method."""

        principals = (
            ('document', 'Document management', (
                ('viewer', 'View all documents', (
                    'document-view',)),
                ('editor', 'Edit any document', (
                    'document-view', 'document-edit')))),)
        self.configurator.set_root_factory(DummyRootFactory)
        root_factory = self.configurator.introspector.get(
            'root factories', None).get('factory')
        self.assertEqual(len(root_factory.__acl__), 2)
        update_acl(self.configurator, principals)
        self.assertEqual(len(root_factory.__acl__), 4)
        self.assertEqual(root_factory.__acl__[2][0], Allow)
        self.assertIn(
            root_factory.__acl__[2][1], ('document.viewer', 'document.editor'))
        self.assertIn('document-view', root_factory.__acl__[2][2])


# =============================================================================
class ULibConfigConfigGet(ConfigTestCase):
    """Unit test class for :func:`lib.config.config_get`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.config.config_get]"""
        from ..lib.config import config_get

        self.assertEqual(config_get(
            self.config, 'app:main', 'site.uid'), 'testchrysalio')
        self.assertEqual(config_get(self.config, 'app:main', 'foo'), None)
        self.assertEqual(
            config_get(self.config, 'app:main', 'foo', 'bar'), 'bar')


# =============================================================================
class ULibConfigConfigGetList(ConfigTestCase):
    """Unit test class for :func:`lib.config.config_get_list`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.config.config_get_list]"""
        from ..lib.config import config_get_list

        self.assertEqual(
            config_get_list(self.config, 'app:main', 'site.uid'),
            ['testchrysalio'])
        self.assertEqual(config_get_list(self.config, 'app:main', 'foo'), [])
        self.assertEqual(
            config_get_list(self.config, 'app:main', 'foo', 'bar'), 'bar')
        self.assertEqual(
            config_get_list(self.config, 'loggers', 'keys'),
            ['root', 'chrysalio', 'sqlalchemy'])


# =============================================================================
class ULibConfigConfigGetNamespace(ConfigTestCase):
    """Unit test class for :func:`lib.config.config_get_namespace`."""

    # -------------------------------------------------------------------------
    def test_unknown_section(self):
        """[u:lib.config.config_get_namespace] unknown section"""
        from ..lib.config import config_get_namespace

        theme = config_get_namespace(self.config, 'foo', 'theme')
        self.assertEqual(len(theme), 0)

    # -------------------------------------------------------------------------
    def test_existing_section(self):
        """[u:lib.config.config_get_namespace] existing section"""
        from ..lib.config import config_get_namespace

        theme = config_get_namespace(self.config, 'app:main', 'theme')
        self.assertIn('roots', theme)
        self.assertIn('patterns', theme)
        self.assertEqual(theme['patterns'], 'Default')


# =============================================================================
class ULibConfigSettingsGetList(TestCase):
    """Unit test class for :func:`lib.config.settings_get_list`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.config.settings_get_list]"""
        from ..lib.config import settings_get_list

        settings = {
            'option1': '',
            'option2': 'test',
            'option3': 'test1, test2'}
        self.assertEqual(settings_get_list(settings, 'option1'), [])
        self.assertEqual(settings_get_list(settings, 'optionX'), [])
        self.assertEqual(
            settings_get_list(settings, 'option1', ['vide']), ['vide'])
        self.assertEqual(
            settings_get_list(settings, 'optionX', ['vide']), ['vide'])
        self.assertEqual(
            settings_get_list(settings, 'option2'), ['test'])
        self.assertEqual(
            settings_get_list(settings, 'option3'), ['test1', 'test2'])


# =============================================================================
class ULibConfigSettingsGetNamespace(TestCase):
    """Unit test class for :func:`lib.config.settings_get_namespace`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.config.settings_get_namespace]"""
        from ..lib.config import settings_get_namespace

        settings = {
            'processor.roots': 'cioprocessor:Processors',
            'build.develop': 'true',
            'build.root': '%(here)s/Builds',
            'build.foo.bar': 'baz'}
        builds = settings_get_namespace(settings, 'build')
        self.assertIsInstance(builds, dict)
        self.assertIn('develop', builds)
        self.assertIn('foo_bar', builds)
        self.assertNotIn('processor.roots', builds)
        self.assertNotIn('roots', builds)


# =============================================================================
class ULibConfigSettingsGetDirectories(TestCase):
    """Unit test class for :func:`lib.config.settings_get_directories`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.config.settings_get_directories]"""
        from ..lib.config import settings_get_directories

        settings = {
            'theme.roots': 'chrysalio:Themes, chrysalio:Foo, chrysalio:Static',
            'theme.patterns':  'Default, Alternative*'}
        directories = settings_get_directories(settings, 'theme', 'theme.conf')
        self.assertIn('Default', directories)
        self.assertIn('Alternative', directories)



"""Setup for Chrysalio."""

from sys import version_info
from os.path import dirname, abspath, join
from setuptools import setup, find_packages

VERSION = '1.0'

INSTALL_REQUIRES = [
    'future',
    'pyramid >= 1.10',
    'pyramid_beaker',
    'pyramid_retry',
    'pyramid_tm',
    'pyramid_chameleon',
    'pyramid_debugtoolbar',
    'SQLAlchemy',
    'zope.sqlalchemy',
    'lxml',
    'bcrypt',
    'pycryptodome',
    'colander',
    'WebHelpers2',
    'docutils',
    'pytz',
    'gitpython',
    'ldap3',
    'Babel',
    'waitress',
]

MODULES_REQUIRE = [
    'ldap3 >= 2.2',
]

TESTS_REQUIRE = MODULES_REQUIRE + [
    'WebTest >= 2.0',
    'testfixtures',
    'flake8',
    'pytest >= 3.7.4',
    'pytest-flake8',
    'pytest-cov',
    'tox',
]

DOCUMENTATION_REQUIRES = [
    'Sphinx',
    'sphinx_rtd_theme',
]

DEVELOPMENT_REQUIRES = TESTS_REQUIRE + DOCUMENTATION_REQUIRES + [
    'pylint',
    'twine',
]

PYTHON_VERSION = version_info[:2]
if PYTHON_VERSION < (3, 5):
    raise RuntimeError('Pyramid requires Python 3.5 or better')

HERE = abspath(dirname(__file__))
with open(join(HERE, 'README.rst')) as hdl:
    README = hdl.read()
with open(join(HERE, 'CHANGES.rst')) as hdl:
    CHANGES = hdl.read()


setup(
    name='Chrysalio',
    version=VERSION,
    description='SDK to build Pyramid Web site with user management',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Framework :: Pyramid',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    ],
    author='Patrick PIERRE',
    author_email='patrick.pierre@lap2.fr',
    url='https://docs.chrysal.io',
    keywords='web wsgi pylons pyramid Chrysalio',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=INSTALL_REQUIRES,
    extras_require={
        'modules': MODULES_REQUIRE,
        'testing': TESTS_REQUIRE,
        'documentation': DOCUMENTATION_REQUIRES,
        'development': DEVELOPMENT_REQUIRES,
    },
    entry_points={
        'paste.app_factory': [
            'main = chrysalio:main',
        ],
        'console_scripts': [
            'ciopopulate = chrysalio.scripts.ciopopulate:main',
            'ciobackup = chrysalio.scripts.ciobackup:main',
            'cioupdate = chrysalio.scripts.cioupdate:main',
        ],
    },
)



"""Profile management view callables."""

from sqlalchemy import desc
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound

from ..lib.i18n import _
from ..lib.log import log_info
from ..lib.form import get_action, Form
from ..lib.filter import Filter
from ..lib.paging import PAGE_SIZES, Paging
from ..lib.utils import make_id
from ..lib.tabset import Tabset
from ..models.populate import xml2db, db2web, web2db
from ..models.dbprofile import DBProfile, DBProfilePrincipal
from . import BaseView


# =============================================================================
class ProfileView(BaseView):
    """Class to manage profile views.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    _DBProfile = DBProfile
    _xml2db = (xml2db,)

    # -------------------------------------------------------------------------
    @view_config(
        route_name='profile_index',
        renderer='chrysalio:Templates/profile_index.pt',
        permission='profile-view')
    @view_config(route_name='profile_index', renderer='json', xhr=True)
    def index(self):
        """List all profiles."""
        # Ajax
        i_creator = self._request.has_permission('profile-create')
        if self._request.is_xhr:
            if i_creator:
                self._import_profiles()
            return {}

        # Action
        action, items = get_action(self._request)
        if action[:4] == 'del!' and i_creator:
            self._delete_profiles(items)
        elif action == 'imp!' and i_creator:
            self._import_profiles()
        elif action[:4] == 'exp!':
            action = self._profiles2response(items)
            if action:
                return action

        # Filter
        paging_id = 'profiles'
        pfilter = Filter(
            self._request, paging_id, (
                ('profile_id', _('Identifier'), False, ''),
                ('i18n_label', _('Label'), False, '')),
            remove=action[:4] == 'crm!' and action[4:] or None)

        # Paging
        defaults = Paging.params(self._request, paging_id, '+profile_id')
        dbquery = pfilter.sql(
            self._request.dbsession.query(self._DBProfile), 'profiles')
        oby = getattr(self._DBProfile, defaults['sort'][1:])
        dbquery = dbquery.order_by(
            desc(oby) if defaults['sort'][0] == '-' else oby)
        profile_paging = Paging(self._request, paging_id, dbquery, defaults)
        profile_paging.set_current_ids('profile_id')

        # Form & completed action
        form = Form(self._request, defaults=defaults)
        form.forget('filter_value')
        if action and action[3] == '!':
            action = ''

        # Breadcrumbs
        self._request.breadcrumbs(_('Profiles'), 2)

        return {
            'action': action, 'items': items, 'form': form, 'pfilter': pfilter,
            'paging': profile_paging, 'PAGE_SIZES': PAGE_SIZES,
            'download_max_size': self._request.registry['settings'][
                'download-max-size'],
            'i_creator': i_creator,
            'i_editor': self._request.has_permission('profile-edit')}

    # -------------------------------------------------------------------------
    @view_config(route_name='profile_index_filter', renderer='json', xhr=True)
    def index_filter(self):
        """Return a list to autocomplete a filter field."""
        return Filter.sql_autocomplete(self._request, self._DBProfile)

    # -------------------------------------------------------------------------
    @view_config(
        route_name='profile_view',
        renderer='chrysalio:Templates/profile_view.pt',
        permission='profile-view')
    def view(self):
        """Show profile."""
        dbprofile = self._get_profile()

        action = get_action(self._request)[0]
        if action == 'exp!':
            action = self._profiles2response((dbprofile.profile_id,))
            if action:
                return action

        # Breadcrumbs
        label = dbprofile.label(self._request)
        self._request.breadcrumbs(
            _('Profile "${l}"', {'l': label}),
            replace=self._request.route_path(
                'profile_edit', profile_id=dbprofile.profile_id))

        return {
            'form': Form(self._request), 'label': label,
            'tabset': Tabset(
                self._request, 'tabProfile',
                dbprofile.settings_tabs(self._request)),
            'dbprofile': dbprofile,
            'navigator': Paging.navigator(
                self._request, 'profiles', dbprofile.profile_id,
                self._request.route_path('profile_view', profile_id='_ID_')),
            'i_editor': self._request.has_permission('profile-edit')}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='profile_create',
        renderer='chrysalio:Templates/profile_edit.pt',
        permission='profile-create')
    @view_config(
        route_name='profile_edit',
        renderer='chrysalio:Templates/profile_edit.pt',
        permission='profile-edit')
    def edit(self):
        """Create or edit a profile."""
        dbprofile = self._get_profile() \
            if 'profile_id' in self._request.matchdict else None
        form = Form(
            self._request,
            *self._DBProfile.settings_schema(self._request, dbprofile),
            obj=dbprofile)

        action = get_action(self._request)[0]
        if action == 'sav!' and form.validate():
            dbprofile = self._save(dbprofile, form.values)
            if dbprofile is not None:
                if 'profile_id' not in self._request.matchdict:
                    self._request.breadcrumbs.pop()
                log_info(
                    self._request,
                    'profile_id' in self._request.matchdict and
                    'profile_edit' or 'profile_create', dbprofile.profile_id)
                return HTTPFound(self._request.route_path(
                    'profile_view', profile_id=dbprofile.profile_id))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        # Breadcrumbs
        label = dbprofile and dbprofile.label(self._request)
        if not dbprofile:
            self._request.breadcrumbs(_('Profile Creation'))
        else:
            self._request.breadcrumbs(
                _('Profile "${l}" Edition', {'l': label}),
                replace=self._request.route_path(
                    'profile_view', profile_id=dbprofile.profile_id))
        return {
            'form': form, 'dbprofile': dbprofile or self._DBProfile,
            'label': label, 'tabset': Tabset(
                self._request, 'tabProfile',
                self._DBProfile.settings_tabs(self._request))}

    # -------------------------------------------------------------------------
    def _get_profile(self):
        """Return the SqlAlchemy object of the selected profile or raise
        an HTTPNotFound exception.

        :rtype: .models.dbprofile.DBProfile
        """
        profile_id = self._request.matchdict['profile_id']
        dbprofile = self._request.dbsession.query(self._DBProfile).filter_by(
            profile_id=profile_id).first()
        if dbprofile is None:
            raise HTTPNotFound()
        return dbprofile

    # -------------------------------------------------------------------------
    def _delete_profiles(self, profile_ids):
        """Delete profiles.

        :param list profile_ids:
            List of profile IDs to delete.
        """
        deleted = []
        for dbprofile in self._request.dbsession.query(self._DBProfile).filter(
                self._DBProfile.profile_id.in_(profile_ids)):
            deleted.append(dbprofile.profile_id)
            self._request.dbsession.delete(dbprofile)
        if deleted:
            log_info(self._request, 'profile_delete', ' '.join(deleted))

    # -------------------------------------------------------------------------
    def _import_profiles(self):
        """Import profiles."""
        # Get current IDs
        profile_ids = {k[0] for k in self._request.dbsession.query(
            DBProfile.profile_id)}

        # Update database
        web2db(self._request, self._xml2db[0], 'profile')

        # Get new IDs
        profile_ids = {k[0] for k in self._request.dbsession.query(
            DBProfile.profile_id)} - profile_ids
        if profile_ids:
            log_info(self._request, 'profile_import', ' '.join(profile_ids))

    # -------------------------------------------------------------------------
    def _profiles2response(self, profile_ids):
        """Export profiles as an XML file embedded in a Pyramid response.

        :param list profile_ids:
            List of profile IDs to export.
        :rtype: :class:`pyramid.response.Response` or ``''``
        """
        dbitems = tuple(self._request.dbsession.query(self._DBProfile).filter(
            self._DBProfile.profile_id.in_(profile_ids)).order_by(
                'profile_id'))
        if not dbitems:
            return ''
        filename = '{0}.{1}.xml'.format(
            len(dbitems) == 1 and dbitems[0].profile_id or
            make_id(self._request.registry['settings']['title'], 'token'),
            self._DBProfile.suffix)

        log_info(
            self._request, 'profile_export',
            ' '.join([k.profile_id for k in dbitems]))
        return db2web(self._request, dbitems, filename)

    # -------------------------------------------------------------------------
    def _save(self, dbprofile, values):
        """Save a profile.

        :type  dbprofile: .models.dbprofile.DBProfile
        :param dbprofile:
            Profile to save.
        :param dict values:
            Form values.
        :rtype: :class:`~.models.dbprofile.DBProfile` instance or ``None``
        """
        creation = dbprofile is None
        dbprofile = dbprofile or self._DBProfile()

        # Update profile
        record = {k: values[k] for k in values if not k.startswith('pcpl:')}
        if not creation:
            record['profile_id'] = dbprofile.profile_id
        error = dbprofile.record_format(record)
        if error:  # pragma: nocover
            self._request.session.flash(error, 'alert')
            return None
        for field in record:
            if getattr(dbprofile, field) != record[field]:
                setattr(dbprofile, field, record[field])

        # Save
        if creation:
            try:
                self._request.dbsession.add(dbprofile)
                self._request.dbsession.flush()
            except (IntegrityError, FlushError):
                self._request.session.flash(
                    _('This profile already exists.'), 'alert')
                return None

        # Update principals
        profile_principals = [k.principal for k in dbprofile.principals]
        for group in self._request.registry['principals']:
            for principal in group[2]:
                principal = '{0}.{1}'.format(group[0], principal[0])
                value = values.get('pcpl:{0}'.format(principal))
                if value and principal not in profile_principals:
                    dbprofile.principals.append(
                        DBProfilePrincipal(principal=principal))
                elif not value and principal in profile_principals:
                    self._request.dbsession.delete(
                        self._request.dbsession.query(DBProfilePrincipal)
                        .filter_by(
                            profile_id=dbprofile.profile_id,
                            principal=principal).first())

        return dbprofile



"""XML manipulation."""

from logging import getLogger
from os.path import join, dirname, basename
from io import BytesIO
from textwrap import wrap
from filecmp import cmpfiles
from shutil import copy2
from json import loads

from lxml import etree

from pyramid.asset import abspath_from_asset_spec

from .i18n import _, translate
from .utils import normalize_spaces


LOG = getLogger(__name__)
XML_NS = '{http://www.w3.org/XML/1998/namespace}'


# =============================================================================
def load_xml(filename, relaxngs=None, data=None, noline=False, parser=None,
             xinclude=False):
    """Load an XML document and validate it against a Relax NG file.

    :param str filename:
        Path to XML file.
    :param dict relaxngs: (optional)
        Relax NG dictionary such as ``{<pattern>: <relax_ng_file>,...}``. If it
        is ``None``, no validation is performed.
    :type data: str, bytes or :class:`lxml.etree.ElementTree`
    :param data: (optional)
        Content of the XML document. If it is not ``None``, it is used in place
        of the content of the file ``filename``.
    :param bool noline: (default=False)
        If ``True``, the error message does not contain line numbers.
    :type parser: :class:`etree.XMLParser`
    :param parser: (optional)
        Specific parser for ``etree.parse`` function.
    :param bool xinclude: (default=False)
        If ``True``, activate XInclude.
    :rtype: str, :class:`TranslationString` or :class:`ElementTree`
    :return:
        An error message or an instance of :class:`lxml.etree.ElementTree`
        class.
    """
    # Read file
    # pylint: disable = protected-access
    if data is None or not isinstance(data, etree._ElementTree):
        if data is not None and not isinstance(data, bytes):
            data = bytes(data.encode())
        try:
            tree = etree.parse(
                filename if data is None else BytesIO(data), parser=parser)
        except IOError:
            return _('Unknown file "${n}"', {'n': basename(filename)})
        except etree.XMLSyntaxError as error:
            return str(error)
    else:
        tree = data
    # pylint: enable = protected-access

    # XInclude
    if xinclude:
        try:
            tree.xinclude()
        except etree.XIncludeError as error:
            return str(error)

    # Validate
    if relaxngs is None:
        return tree
    error = validate_xml(tree, relaxngs, noline)
    if error is not None:
        return error

    return tree


# =============================================================================
def validate_xml(tree, relaxngs, noline=False):
    """Load an XML document and validate it against a Relax NG file.

    :type  tree: lxml.etree.ElementTree
    :param tree:
        XML document.
    :param dict relaxngs:
        Relax NG dictionary such as ``{<pattern>: <relax_ng_file>,...}``.
    :param bool noline: (default=False)
        If ``True``, the error message does not contain line numbers.
    :rtype: str, :class:`TranslationString` or ``None``
    :return:
        An error message or ``None``.
    """
    # Find the right RelaxNG
    relaxng = None
    pattern = None
    root = tree.getroot()
    for pattern in relaxngs:
        chunks = pattern.split(',')
        chunk = chunks[0].strip()
        if root.tag != chunk:
            continue
        for chunk in chunks[1:]:
            chunk = chunk.strip().split('=')
            if root.get(chunk[0]) != chunk[1]:
                chunk = None
                break
        if chunk is not None:
            relaxng = relaxngs[pattern]
            break
    if relaxng is None:
        return _('${tag}: Relax NG not found', {'tag': tree.getroot().tag})

    # Load Relax NG
    if isinstance(relaxng, str):
        try:
            relaxng = etree.RelaxNG(etree.parse(relaxng))
        except IOError as error:
            return str(error)
        except (etree.XMLSyntaxError, etree.RelaxNGParseError) as error:
            return '"{0}": {1}'.format(relaxng, error)
        relaxngs[pattern] = relaxng

    # Validate
    if not relaxng.validate(tree):
        error = relaxng.error_log.last_error
        return error.message if noline else \
            _('Line ${l}: ${m}', {'l': error.line, 'm': error.message})
    return None


# =============================================================================
def relaxng4validation(relaxng, attributes=('version',)):
    """Transform a Relax NG dictionary with keys ``'root'``, ``'file'`` and
    possibly ``'namespace'`` and ``'version'`` into a dictionary compatible
    with :func:`validate_xml`.

    :param dict relaxng:
        A Chrysalio Relax NG dictionary.
    :param tuple attributes: (default=('version',))
        Attributes to take into account in the pattern.
    :rtype: :class:`dict` or ``None``
    """
    if not relaxng:
        return None
    if 'namespace' in relaxng:
        pattern = '{{{0}}}{1}'.format(
            relaxng['namespace'], relaxng['root'])
    else:
        pattern = relaxng['root']

    for attribute in attributes:
        if attribute in relaxng:
            pattern += ', {0}={1}'.format(attribute, relaxng[attribute])

    return {pattern: relaxng['file']}


# =============================================================================
def load_xslt(filename):
    """Load a XSL file and create a etree.XSLT object.

    :rtype: :class:`lxml.etree.XSLT` or :class:`str`
    """
    try:
        xslt = etree.XSLT(etree.parse(filename))
    except (IOError, etree.XSLTParseError, etree.XMLSyntaxError) as error:
        return str(error)
    return xslt


# =============================================================================
def create_entire_xml(relaxng, elements, validation=True):
    """Create an entire XML document composed of all ``elements``.

    :param dict relaxng:
        A dictionary with the name of the root element, the value of attribute
        version and path to the Relax NG file to validate the result.
    :param list elements:
        A list of :class:`lxml.etree._Element` objects.
    :param bool validation: (default=True)
        Validate the result.
    :rtype: :class:`~pyramid.i18n.TranslationString` or
        :class:`lxml.etree._Element`
    :return:
        An error message or an object :class:`lxml.etree._Element`.
    """
    def _label(elt):
        """Get label or name of ``elt``."""
        label = elt.get('{0}id'.format(XML_NS)) or elt.get('id') \
            or (elt.findtext('login') is not None and elt.findtext('login')) \
            or (elt.findtext('label') is not None and elt.findtext('label')) \
            or (elt.findtext('title') is not None and elt.findtext('title')) \
            or '~'
        return normalize_spaces(label)

    # Create XML root
    if 'namespace' in relaxng:
        root = '{{{0}}}{1}'.format(relaxng['namespace'], relaxng['root'])
        root_elt = etree.Element(
            root, version=relaxng['version'],
            nsmap={None: relaxng['namespace']})
    else:
        root = relaxng['root']
        root_elt = etree.Element(relaxng['root'], version=relaxng['version'])

    # Single export
    if len(elements) == 1 and 'namespace' not in relaxng:
        root_elt.append(elements[0])
    # Multiple export or namepsace
    else:
        tag = None
        for elt in elements:
            if tag != elt.tag:
                tag = elt.tag
                if tag.endswith('s'):
                    root_elt.append(etree.Comment(' {0} '.format('=' * 68)))
                    root_elt.append(etree.Comment('{0:^70}'.format(tag)))
                    root_elt.append(etree.Comment(' {0} '.format('=' * 68)))
                    group_elt = root_elt
                else:
                    root_elt.append(etree.Comment(' {0} '.format('=' * 68)))
                    root_elt.append(etree.Comment('{0:^70}'.format(
                        '{0}s'.format(tag))))
                    root_elt.append(etree.Comment(' {0} '.format('=' * 68)))
                    group_elt = etree.SubElement(root_elt, '{0}s'.format(tag))
            if tag.endswith('s'):
                group_elt.append(elt)
            else:
                group_elt.append(etree.Comment(' {0:~^66} '.format(
                    ' {0} '.format(_label(elt)))))
                group_elt.append(elt)

    # Validate the result
    error = None
    if validation:
        if 'namespace' in relaxng:
            root_elt = etree.XML(etree.tostring(root_elt, encoding='utf-8'))
        error = validate_xml(
            etree.ElementTree(root_elt), relaxng4validation(relaxng), True)

    return root_elt if error is None else translate(error)


# =============================================================================
def i18n_xml_text(root_elt, xpath, namespaces=None):
    """Return a dictionary with the localized texts contained in an XML
    element.

    :type  root_elt: lxml.etree.Element
    :param root_elt:
        XML root element of the localized text.
    :param str xpath:
        XPath expression to select localized texts.
    :param dict namespaces: (optional)
        Dictionary of possible name spaces.
    :rtype: dict
    :return:
        A dictionary such ``{lang1: text_in_lang1,...}``
    """
    i18n = {}
    for elt in root_elt.xpath(xpath, namespaces=namespaces):
        lang = elt.get('{0}lang'.format(XML_NS))
        if lang and elt.text:
            i18n[lang] = normalize_spaces(elt.text)
    return i18n


# =============================================================================
def db2xml_i18n_labels(dbitem, root_elt, depth):
    """Serialize i18n label and descriptio.

    :param dbitem:
        SQLAlchemy item.
    :type  root_elt: lxml.etree.Element
    :param root_elt:
        Root XML element.
    :param int depth:
        Depth of the parent element in the entire XML structure.
    """
    i18n = loads(dbitem.i18n_label)
    for lang in sorted(i18n):
        elt = etree.SubElement(root_elt, 'label')
        elt.set('{0}lang'.format(XML_NS), lang)
        elt.text = i18n[lang]

    if hasattr(dbitem, 'i18n_description'):
        i18n = dbitem.i18n_description
        if i18n:
            for lang in sorted(i18n):
                elt = etree.SubElement(root_elt, 'description')
                elt.set('{0}lang'.format(XML_NS), lang)
                elt.text = xml_wrap(i18n[lang], depth)


# =============================================================================
def xml_wrap(text, depth):
    """Wrap a text according to the depth of the parent element.

    :param str text:
        Text to wrap and indent.
    :param int depth:
        Depth of the parent element in the entire XML structure.
    :rtype: str
    """
    indent = ' ' * 2 * (depth + 1)
    return '\n{0}\n{1}'.format(
        '\n'.join(
            ['{0}{1}'.format(indent, k)
             for k in wrap(text, 79 - 2 * (depth + 1))]), ' ' * 2 * depth)


# =============================================================================
def check_chrysalio_rng(relaxng_dir):
    """Check if the Relax NG files in ``relaxng_dir`` directory is the last
    version and possibly update them.

    :param str relaxng_dir:
        Directory for Relax NG files.
    """
    relaxng_dir = abspath_from_asset_spec(relaxng_dir)
    cio_relaxng_dir = join(dirname(__file__), '..', 'RelaxNG')
    for name in cmpfiles(cio_relaxng_dir, relaxng_dir,
                         ('chrysalio.rnc', 'chrysalio.rng'))[1]:
        try:
            copy2(join(cio_relaxng_dir, name), relaxng_dir)
        except IOError:  # pragma: nocover
            LOG.warning('"%s" is not up to date.', name)



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``includes.modules.views`` classes."""

from ....tests import DBUnitTestCase


# =============================================================================
class UIncludesModulesViewsModulesView(DBUnitTestCase):
    """Unit test class for testing
    :class:`includes.modules.views.ModulesView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ....security import PRINCIPALS
        super(UIncludesModulesViewsModulesView, self).setUp()

        self.configurator.add_route('home', '/home')
        self.configurator.add_route('modules_view', '/modules/view')
        self.configurator.add_route('modules_edit', '/modules/edit')
        self.request.registry['principals'] = list(PRINCIPALS)

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:includes.modules.views.ModulesView.view]"""
        from collections import namedtuple
        from ..views import ModulesView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='modules_view')

        response = ModulesView(self.request).view()
        self.assertIn('form', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Modules')

    # -------------------------------------------------------------------------
    def test_edit(self):
        """[u:includes.modules.views.ModulesView.edit]"""
        from collections import namedtuple, OrderedDict
        from ..views import ModulesView

        self.request.matched_route = namedtuple('Route', 'name')(
            name='modules_edit')
        self.request.registry['modules'] = OrderedDict()
        self.request.registry['modules_off'] = set()

        response = ModulesView(self.request).edit()
        self.assertIn('form', response)
        self.assertIn('dbmodules', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Modules Edition')

    # -------------------------------------------------------------------------
    def test_edit_save(self):
        """[u:includes.modules.views.ModulesView.edit] save"""
        from collections import OrderedDict
        from ....tests import TEST_INI
        from ....modules.cioskeleton import ModuleCioSkeleton
        from ..models import DBModule
        from ..views import ModulesView

        self.request.registry['modules'] = OrderedDict((
            ('chrysalio.modules.cioskeleton', ModuleCioSkeleton(TEST_INI)),))
        self.request.registry['modules_off'] = set(
            ['chrysalio.modules.cioskeleton'])
        self.request.POST = {
            'chrysalio.modules.cioskeleton': True, 'sav!.x': True}

        ModulesView(self.request).edit()
        dbmodule = self.request.dbsession.query(DBModule).first()
        self.assertIsNone(dbmodule)
        self.assertNotIn(
            'chrysalio.modules.cioskeleton',
            self.request.registry['modules_off'])

    # -------------------------------------------------------------------------
    def test_save(self):
        """[u:includes.modules.views.ModulesView._save]"""
        from collections import OrderedDict
        from ....tests import TEST_INI
        from ....modules import Module
        from ..views import ModulesView

        self.request.POST = {'sav!.x': True}
        self.request.session['menu'] = ()
        self.request.session['modes'] = ['home', 'Home', None]
        self.request.registry['modules'] = OrderedDict()
        self.request.registry['modules_off'] = set()

        self.request.POST['module1'] = True
        ModulesView(self.request).edit()
        self.assertEqual(len(self.request.registry['modules_off']), 0)

        module1 = Module(TEST_INI)
        module1.name = 'Module 1'
        module2 = Module(TEST_INI)
        module2.name = 'Module 2'
        module2.dependencies = ('module1',)
        module3 = Module(TEST_INI)
        module3.name = 'Module 3'
        module3.dependencies = ('module2',)
        self.request.registry['modules'] = OrderedDict((
            ('module1', module1), ('module2', module2), ('module3', module3)))
        self.request.registry['modules_off'] = set(
            ['module1', 'module2', 'module3'])

        self.request.POST['module1'] = True
        ModulesView(self.request).edit()
        self.assertEqual(len(self.request.registry['modules_off']), 2)
        self.assertNotIn('module1', self.request.registry['modules_off'])
        self.assertIn('module2', self.request.registry['modules_off'])
        self.assertIn('module3', self.request.registry['modules_off'])

        self.request.POST['module1'] = False
        self.request.POST['module2'] = True
        ModulesView(self.request).edit()
        self.assertEqual(len(self.request.registry['modules_off']), 1)
        self.assertNotIn('module1', self.request.registry['modules_off'])
        self.assertNotIn('module2', self.request.registry['modules_off'])
        self.assertIn('module3', self.request.registry['modules_off'])

        self.request.POST['module1'] = False
        self.request.POST['module2'] = False
        self.request.POST['module3'] = True
        ModulesView(self.request).edit()
        self.assertEqual(len(self.request.registry['modules_off']), 0)

        self.request.POST['module1'] = False
        self.request.POST['module2'] = True
        self.request.POST['module3'] = True
        ModulesView(self.request).edit()
        self.assertEqual(len(self.request.registry['modules_off']), 3)



"""CioSkeleton secutity functionalities."""

from ...lib.i18n import _
from ...lib.config import update_acl


PRINCIPALS_CIOSKELETON = (
    ('mode', _('Mode management'), (
        ('skeleton', _('Access to skeleton mode'), ('mode-skeleton',)),
    )),
    ('bone', _('Bone management'), (
        ('viewer', _('View bone'), ('bone-view',)),)))


# =============================================================================
def includeme(configurator):
    """Function to include security.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    update_acl(configurator, PRINCIPALS_CIOSKELETON)



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.filter`` class methods."""

from json import dumps
from unittest import TestCase

from pyramid.testing import DummyRequest
from pyramid.httpexceptions import HTTPForbidden

from ..models.dbgroup import DBGroup
from . import DBUnitTestCase


# =============================================================================
class ULibFilterFilter(TestCase):
    """Unit test class for :class:`lib.filter.Filter`."""

    # -------------------------------------------------------------------------
    @classmethod
    def make_one(cls, request):
        """Create an instance of Filter.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        rtype: Filter
        """
        from ..lib.filter import Filter

        return Filter(
            request, 'users', (
                ('login', 'Login', True, None),
                ('name', 'Name', True, ''),
                ('active', 'Active', False, True),
                ('age', 'Age', False, 0),
                ('group', 'Group', False, [
                    ('', ' '), ('foo', 'Foo'), ('bar', 'Bar')])),
            (('active', '=', True), ('group', '!=', 'foo')), '1')

    # -------------------------------------------------------------------------
    def test__str__(self):
        """[u:lib.filter.Filter.__str__]"""
        afilter = self.make_one(DummyRequest())
        text = str(afilter)
        self.assertEqual(text, "[[('active', '=', True)]]")

    # -------------------------------------------------------------------------
    def test_is_empty(self):
        """[u:lib.filter.Filter.is_empty]"""
        afilter = self.make_one(DummyRequest())
        self.assertFalse(afilter.is_empty())

        afilter.remove_condition('0')
        self.assertTrue(afilter.is_empty())

    # -------------------------------------------------------------------------
    def test_clear(self):
        """[u:lib.filter.Filter.clear]"""
        request = DummyRequest()
        afilter = self.make_one(request)
        afilter.clear()
        self.assertTrue(afilter.is_empty())
        self.assertIn('filters', request.session)
        self.assertIn('users', request.session['filters'])
        self.assertFalse(len(request.session['filters']['users']))

    # -------------------------------------------------------------------------
    def test_append_condition(self):
        """[u:lib.filter.Filter.append_condition]"""
        afilter = self.make_one(DummyRequest(params={
            'filter': True,
            'filter_key_login': 'login', 'filter_value_login': 'user1',
            'filter_key': 'group', 'filter_value': 'bar'}))
        afilter.remove_condition('0')
        self.assertFalse(afilter.is_empty())
        afilter.remove_condition('1')
        self.assertFalse(afilter.is_empty())
        afilter.remove_condition('0')
        self.assertTrue(afilter.is_empty())

        afilter = self.make_one(DummyRequest())
        afilter.remove_condition('0')
        afilter.append_condition('foo', '=', True)
        self.assertTrue(afilter.is_empty())

        afilter.append_condition('login', '=', '')
        self.assertTrue(afilter.is_empty())

        afilter.append_condition('login', '=', 'user1', 'OR')
        self.assertFalse(afilter.is_empty())

        afilter.append_condition('age', '>', 'foo')
        self.assertIn('<span>Age > 0</span>', afilter.html_conditions())
        afilter.append_condition('age', '<', '18', 'OR')
        self.assertIn('<span>Age < 18</span>', afilter.html_conditions())

    # -------------------------------------------------------------------------
    def test_remove_condition(self):
        """[u:lib.filter.Filter.remove_condition]"""
        afilter = self.make_one(DummyRequest(params={
            'filter': True,
            'filter_key': 'login', 'filter_value': 'user1'}))
        afilter.remove_condition('a')
        self.assertFalse(afilter.is_empty())
        afilter.remove_condition(None)
        self.assertFalse(afilter.is_empty())
        afilter.remove_condition('1')
        self.assertFalse(afilter.is_empty())
        afilter.remove_condition('1')
        self.assertFalse(afilter.is_empty())
        afilter.remove_condition('0')
        self.assertTrue(afilter.is_empty())

    # -------------------------------------------------------------------------
    def test_html_conditions(self):
        """[u:lib.filter.Filter.html_conditions]"""
        afilter = self.make_one(DummyRequest())
        afilter.remove_condition('0')
        self.assertEqual(afilter.html_conditions(), '')

        afilter = self.make_one(DummyRequest())
        afilter.append_condition('login', '=', 'user1')
        afilter.append_condition('login', '=', 'user2', 'OR')
        afilter.append_condition('group', '=', 'foo')
        html = afilter.html_conditions()
        self.assertIn('name="crm!0.x"', html)
        self.assertIn('<span>Active</span>', html)
        self.assertIn('<strong>AND</strong>', html)
        self.assertIn('name="crm!1.x"', html)
        self.assertIn('<span>Login = user1</span>', html)
        self.assertIn('<strong>OR</strong>', html)
        self.assertIn('<span>Login = user2</span>', html)
        self.assertIn('name="crm!2.x"', html)
        self.assertIn('<span>Group = Foo</span>', html)
        self.assertNotIn('name="crm!3.x"', html)

    # -------------------------------------------------------------------------
    def test_html_inputs(self):
        """[u:lib.filter.Filter.html_inputs]"""
        from ..lib.form import Form
        from ..lib.filter import Filter

        request = DummyRequest()
        form = Form(request)
        afilter = Filter(
            request, 'users1', (
                ('login', 'Login', True, None),
                ('active', 'Active', False, True),
                ('group', 'Group', False, [
                    ('', ' '), ('foo', 'Foo'), ('bar', 'Bar')])))

        html = afilter.html_inputs(form)
        self.assertNotIn('id="filter_operator_login"', html)
        self.assertIn('id="filter_comparison_login"', html)
        self.assertIn('id="filter_value_login"', html)
        self.assertNotIn('id="filter_value_active', html)
        self.assertNotIn('id="filter_operator"', html)
        self.assertIn('id="filter_comparison"', html)
        self.assertIn('id="filter_value"', html)

        afilter = self.make_one(request)
        html = afilter.html_inputs(form)
        self.assertNotIn('id="filter_operator_login"', html)
        self.assertIn('id="filter_operator"', html)
        self.assertIn('id="filter_comparison"', html)
        self.assertIn('id="filter_value"', html)

        afilter = Filter(
            request, 'users2', (
                ('login', 'Login', True, None),
                ('name', 'Name', True, ''),
                ('active', 'Active', True, True),
                ('group', 'Group', True, [
                    ('', ' '), ('foo', 'Foo'), ('bar', 'Bar')])))
        html = afilter.html_inputs(form)
        self.assertIn('id="filter_value_login"', html)
        self.assertIn('id="filter_value_active"', html)
        self.assertNotIn('id="filter_value"', html)
        self.assertIn('class="cioAutocomplete"', html)

    # -------------------------------------------------------------------------
    def test_whoosh(self):
        """[u:lib.filter.Filter.whoosh]"""
        from ..lib.filter import Filter

        wfilter = self.make_one(DummyRequest())
        wfilter.append_condition('login', '=', 'user1')
        wfilter.append_condition('login', '=', 'user2', 'OR')

        # Complex query
        fieldnames, wquery = wfilter.whoosh()
        self.assertEqual(len(fieldnames), 2)
        self.assertIn('active', fieldnames)
        self.assertIn('login', fieldnames)
        self.assertEqual(
            wquery, '(active:TRUE) AND (login:(user1) OR login:(user2))')

        # Equal with space
        wfilter = Filter(
            DummyRequest(), 'users', (('name', 'Name', False, ''),),
            (('name', '=', 'John DOE'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(name:"John DOE")')

        # Different
        wfilter = Filter(
            DummyRequest(), 'users', (('login', 'Login', False, None),),
            (('login', '!=', 'user1'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(NOT login:(user1))')

        # Different with space
        wfilter = Filter(
            DummyRequest(), 'users', (('name', 'Name', False, ''),),
            (('name', '!=', 'John DOE'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(NOT name:"John DOE")')

        # Greater
        wfilter = Filter(
            DummyRequest(), 'users', (('login', 'Login', False, None),),
            (('login', '>', 'user1'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(login:{user1 TO})')

        # Greater or equal
        wfilter = Filter(
            DummyRequest(), 'users', (('login', 'Login', False, None),),
            (('login', '>=', 'user1'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(login:[user1 TO])')

        # Lower
        wfilter = Filter(
            DummyRequest(), 'users', (('login', 'Login', False, None),),
            (('login', '<', 'user1'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(login:{TO user1})')

        # Lower or equal
        wfilter = Filter(
            DummyRequest(), 'users', (('login', 'Login', False, None),),
            (('login', '<=', 'user1'),))
        wquery = wfilter.whoosh()[1]
        self.assertEqual(wquery, '(login:[TO user1])')


# =============================================================================
class ULibFilterFilterSql(DBUnitTestCase):
    """Unit test class for SQL method of :class:`lib.filter.Filter` class."""

    # -------------------------------------------------------------------------
    def test_sql(self):
        """[u:lib.filter.Filter.sql]"""
        from ..lib.filter import Filter
        from ..models.dbuser import DBUser

        dbsession = self.request.dbsession
        pfilter = Filter(
            self.request, 'users1',
            (('login', 'Login', False, None),
             ('active', 'Active', False, True)))
        dbquery = pfilter.sql(
            dbsession.query(DBUser.user_id, DBUser.login), 'users')
        self.assertIn('users.user_id AS users_user_id', str(dbquery.params()))

        pfilter = Filter(
            self.request, 'users2',
            (('login', 'Login', False, None),
             ('active', 'Active', False, True)),
            (('login', '=', 'user1'), ('login', '!=', 'user2', 'OR'),
             ('active', '=', True)))
        dbquery = pfilter.sql(
            dbsession.query(DBUser.user_id, DBUser.login), 'users')
        sql = str(dbquery.params())
        self.assertIn('users.login LIKE ? OR users.login NOT LIKE ?', sql)
        self.assertIn('users.active = ?', sql)

        pfilter = Filter(
            self.request, 'users3',
            (('login', 'Login', False, None),
             ('active', 'Active', False, True)),
            (('login', '=', 'user1'), ('active', '=', True, 'OR'),
             ('active', '=', True)))
        dbquery = pfilter.sql(
            dbsession.query(DBUser.user_id, DBUser.login),
            'users', ('active',))
        self.assertNotIn('users.active = ?', str(dbquery.params()))

    # -------------------------------------------------------------------------
    def test_sql_autocomplete(self):
        """[u:lib.filter.Filter.autocomplete]"""
        from ..lib.filter import Filter
        from ..models.dbprofile import DBProfile

        dbsession = self.request.dbsession
        dbsession.add(DBProfile(
            profile_id='profile_manager',
            i18n_label=dumps({'en': 'Profile manager'}),
            i18n_description={'en': 'Permission to create profiles'}))
        dbsession.add(DBProfile(
            profile_id='user_manager',
            i18n_label=dumps({'en': 'User manager'}),
            i18n_description={'en': 'Permission to create users'}))
        dbsession.add(DBGroup(
            group_id='managers', i18n_label=dumps({'en': 'Managers'})))

        # Not an AJAX call
        self.request.is_xhr = False
        self.assertRaises(
            HTTPForbidden, Filter.sql_autocomplete, self.request, DBProfile)

        # Unknown field
        self.request.is_xhr = True
        self.request.params = {'field': 'foo'}
        self.assertRaises(
            HTTPForbidden, Filter.sql_autocomplete, self.request, DBProfile)

        # Field i18n_label
        self.request.params = {'field': 'i18n_label', 'value': 'profile'}
        data = Filter.sql_autocomplete(self.request, DBProfile)
        self.assertIsInstance(data, list)
        self.assertIn('Profile manager', data)

        # Field profile_id with WHERE clause
        data = Filter.sql_autocomplete(
            self.request, DBProfile, where={'profile_id': 'profile_manager'},
            width=7)
        self.assertIn('Profile', data)
        self.request.params = {'field': 'i18n_label', 'value': 'user'}
        data = Filter.sql_autocomplete(
            self.request, DBProfile, where={'profile_id': 'profile_manager'})
        self.assertFalse(data)



# pylint: disable = import-outside-toplevel
"""Tests of ``initialize.Initialize`` class."""

from pkg_resources import register_loader_type, DefaultProvider
from _pytest.assertion.rewrite import AssertionRewritingHook

# pylint: disable = unused-import
from ..models.dbgroup import DBGroup  # noqa
# pylint: enable = unused-import
from . import DBUnitTestCase


# =============================================================================
class UInitializeInitialize(DBUnitTestCase):
    """Unit test class for :class:`initialize.Initialize`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        super(UInitializeInitialize, self).setUp()
        register_loader_type(AssertionRewritingHook, DefaultProvider)

    # -------------------------------------------------------------------------
    def test_complete(self):
        """[u:initialize.Initialize.complete]"""
        from os import makedirs
        from os.path import dirname
        from transaction import manager
        from . import TEST_INI, TEST1_INI, TEST_DIR
        from ..initialize import Initialize

        global_config = {'__file__': TEST1_INI, 'here': dirname(TEST1_INI)}
        self.configurator.get_settings()['temporary'] = TEST_DIR
        makedirs(TEST_DIR)

        self.assertRaises(
            SystemExit, Initialize(self.configurator).complete,
            global_config, 'chrysalio', 'ciopopulate')

        settings = self.configurator.get_settings()
        settings['site.uid'] = 'testchrysalio'
        self.assertRaises(
            SystemExit, Initialize(self.configurator).complete,
            global_config, 'chrysalio', 'ciopopulate')

        global_config = {'__file__': TEST_INI, 'here': dirname(TEST_INI)}
        self.configurator.get_settings()['temporary'] = TEST_DIR
        self.assertRaises(
            SystemExit, Initialize(self.configurator).complete,
            global_config, 'chrysalio', 'ciopopulate')

        manager.commit()
        settings['testing'] = 'true'
        Initialize(self.configurator).complete(
            global_config, 'chrysalio', 'ciopopulate')
        self.assertIn('themes', self.configurator.registry)
        self.assertEqual(self.configurator.registry['settings']['theme'], '')
        self.assertEqual(len(self.configurator.registry['themes']), 1)

    # -------------------------------------------------------------------------
    def test_add_static_views(self):
        """[u:initialize.Initialize.add_static_views]"""
        from ..initialize import Initialize
        from . import TEST_DIR

        url = '/chrysalio/test/'
        self.assertIsNone(
            self.configurator.introspector.get('static views', url))

        Initialize(self.configurator).add_static_views(
            'chrysalio', (('test', TEST_DIR),))
        self.assertIsNotNone(
            self.configurator.introspector.get('static views', url))



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``models`` functions and class."""

from os import makedirs, listdir
from os.path import exists
from json import dumps

from sqlalchemy import Column, String, Integer, Text, PickleType

from . import DBUnitTestCase


# =============================================================================
class UModelsDBBaseClass(DBUnitTestCase):
    """Unit test class for :class:`models.DBBaseClass`."""

    # -------------------------------------------------------------------------
    def test_label(self):
        """[u:models.DBBaseClass.label]"""
        from ..models import DBDeclarativeClass
        from ..models.dbbase import DBBaseClass

        class DBDummy1(DBDeclarativeClass, DBBaseClass):
            """SQLAlchemy-powered dummy model."""
            # pylint: disable = too-few-public-methods
            __tablename__ = 'dummy1'
            uid = Column(Integer, primary_key=True)
            i18n_label = Column(Text(), nullable=False)

        dbitem = DBBaseClass()
        self.assertEqual(dbitem.label(self.request), '')

        dbitem = DBDummy1(
            i18n_label=dumps({
                'en': 'FTP Monitoring', 'fr': 'Surveillance FTP'}))
        self.assertEqual(dbitem.label(self.request), 'FTP Monitoring')

        self.request.session['lang'] = 'fr'
        self.assertEqual(dbitem.label(self.request), 'Surveillance FTP')

    # -------------------------------------------------------------------------
    def test_description(self):
        """[u:models.DBBaseClass.description]"""
        from ..models import DBDeclarativeClass
        from ..models.dbbase import DBBaseClass

        class DBDummy2(DBDeclarativeClass, DBBaseClass):
            """SQLAlchemy-powered dummy model."""
            # pylint: disable = too-few-public-methods
            __tablename__ = 'dummy2'
            uid = Column(Integer, primary_key=True)
            i18n_description = Column(PickleType(1))

        dbitem = DBBaseClass()
        self.assertEqual(dbitem.description(self.request), '')

        dbitem = DBDummy2(
            i18n_description={
                'en': 'Periodically scan a FTP server.',
                'fr': 'Scrute périodiquement un serveur FTP.'})
        self.assertEqual(
            dbitem.description(self.request),
            'Periodically scan a FTP server.')

        self.request.session['lang'] = 'fr'
        self.assertEqual(
            dbitem.description(self.request),
            'Scrute périodiquement un serveur FTP.')

    # -------------------------------------------------------------------------
    def test_attachments2directory(self):
        """[u:models.DBBaseClass.attachments2directory]"""
        from ..models import DBDeclarativeClass, ID_LEN
        from ..models.dbbase import DBBaseClass
        from . import ATTACHMENTS_DIR, TEST_DIR

        class DBDummy3(DBDeclarativeClass, DBBaseClass):
            """SQLAlchemy-powered dummy model."""
            # pylint: disable = too-few-public-methods
            attachments_dir = 'Dummy'
            __tablename__ = 'dummy3'
            uid = Column(Integer, primary_key=True)
            attachments_key = Column(String(ID_LEN + 20))
            picture = Column(String(ID_LEN + 4))

        if not exists(TEST_DIR):
            makedirs(TEST_DIR)

        # No attachment
        dbitem = DBBaseClass()
        dbitem.attachments2directory(ATTACHMENTS_DIR, TEST_DIR)
        self.assertFalse(listdir(TEST_DIR))

        # Incorrect attachment
        dbitem = DBDummy3(
            attachments_key='Test1', picture='testX.svg')
        dbitem.attachments2directory(ATTACHMENTS_DIR, TEST_DIR)
        self.assertFalse(listdir(TEST_DIR))

        # Correct attachment
        dbitem = DBDummy3(
            attachments_key='Test1', picture='test1.svg')
        dbitem.attachments2directory(ATTACHMENTS_DIR, TEST_DIR)
        self.assertTrue(listdir(TEST_DIR))



"""SQLAlchemy-powered model definitions for bones."""

from sqlalchemy import Column
from sqlalchemy import String
from lxml import etree

from ....lib.utils import make_id
from ....lib.i18n import _
from ....models import DBDeclarativeClass, ID_LEN, LABEL_LEN
from ....models.dbbase import DBBaseClass
from ..relaxng import RELAXNG_CIOSKELETON


# =============================================================================
class DBBone(DBDeclarativeClass, DBBaseClass):
    """SQLAlchemy-powered bone class."""

    suffix = 'cioskt'
    attachments_dir = 'cioskt'
    _settings_tabs = (_('Information'),)

    __tablename__ = 'bones'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    bone_id = Column(String(ID_LEN), primary_key=True)
    label = Column(String(LABEL_LEN), nullable=False)
    attachments_key = Column(String(ID_LEN + 20))
    picture = Column(String(ID_LEN + 4))

    # -------------------------------------------------------------------------
    @classmethod
    def xml2db(cls, dbsession, bone_elt, error_if_exists=True, kwargs=None):
        """Load a bone from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  bone_elt: lxml.etree.Element
        :param bone_elt:
            Bone XML element.
        :param bool error_if_exists: (default=True)
            It returns an error if bone already exists.
        :param dict kwargs: (optional)
            Dictionary of keyword arguments.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        # pylint: disable = unused-argument
        # Check if already exists
        bone_id = make_id(bone_elt.get('id'), 'token', ID_LEN)
        dbbone = dbsession.query(cls).filter_by(bone_id=bone_id).first()
        if dbbone is not None:
            if error_if_exists:
                return _('Bone "${s}" already exists.', {'s': bone_id})
            return None

        # Create bone
        record = cls.record_from_xml(bone_id, bone_elt)
        error = cls.record_format(record)
        if error:
            return error
        dbsession.add(cls(**record))
        return None

    # -------------------------------------------------------------------------
    @classmethod
    def record_from_xml(cls, bone_id, bone_elt):
        """Convert a bone XML element into a dictionary.

        :param str bone_id:
            Bone ID.
        :type  bone_elt: lxml.etree.Element
        :param bone_elt:
            Bone XML element.
        :rtype: dict
        """
        namespace = RELAXNG_CIOSKELETON['namespace']
        attachments = bone_elt.find('{{{0}}}attachments'.format(namespace))
        return {
            'bone_id': bone_id,
            'label': bone_elt.findtext(
                '{{{0}}}label'.format(namespace)) or bone_id,
            'attachments_key': attachments is not None and attachments.get(
                'key') or None,
            'picture': attachments is not None and attachments.findtext(
                '{{{0}}}picture'.format(namespace)) or None}

    # -------------------------------------------------------------------------
    @classmethod
    def record_format(cls, record):
        """Check and possibly correct a record before inserting it in the
        database.

        :param dict record:
            Dictionary of values to check.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        for k in [i for i in record if record[i] is None]:
            del record[k]

        # Bone ID
        if not record.get('bone_id'):
            return _('Bone without ID.')
        record['bone_id'] = make_id(record['bone_id'], 'token', ID_LEN)
        return None

    # -------------------------------------------------------------------------
    def db2xml(self):
        """Serialize a bone to a XML representation.

        :rtype: lxml.etree.Element
        """
        bone_elt = etree.Element('{{{0}}}bone'.format(
            RELAXNG_CIOSKELETON['namespace']))
        bone_elt.set('id', self.bone_id)
        etree.SubElement(bone_elt, 'label').text = self.label
        if self.attachments_key:
            elt = etree.SubElement(
                bone_elt, 'attachments', key=self.attachments_key)
            if self.picture:
                etree.SubElement(elt, 'picture').text = self.picture

        return bone_elt



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioskeleton.views.bone.BoneView`` class."""

from collections import namedtuple

from pyramid.httpexceptions import HTTPForbidden

from ....tests import DBUnitTestCase


# =============================================================================
class UViewsBoneBoneView(DBUnitTestCase):
    """Unit test class for testing
 :class:`modules.cioskeleton.views.bone.BoneView`."""

    # -------------------------------------------------------------------------
    def test_view(self):
        """[u:modules.cioskeleton.views.bone.BoneView.index]"""
        from ....tests import TEST_INI
        from ..views.bone import BoneView
        from .. import ModuleCioSkeleton

        # Inactive module
        self.request.matched_route = namedtuple('Route', 'name')(
            name='bone_index')
        self.assertRaises(HTTPForbidden, BoneView(self.request).index)

        # Active module
        self.request.registry['modules'] = {
            'chrysalio.modules.cioskeleton': ModuleCioSkeleton(TEST_INI)}
        self.request.registry['modules_off'] = set()
        self.assertFalse(BoneView(self.request).index())
